"""HEC-HMS Stac Item asset classes."""

from pystac import MediaType
import re
from hecstac.common.asset_factory import GenericAsset
from hecstac.hms.parser import (
    BasinFile,
    ControlFile,
    GageFile,
    GridFile,
    MetFile,
    PairedDataFile,
    ProjectFile,
    RunFile,
    SqliteDB,
    TerrainFile,
)

HMS_VERSION = "hms:version"
HMS_TITLE = "hms:title"
HMS_DESCRIPTION = "hms:description"
HMS_UNIT_SYSTEM = "hms:unit_system"
HMS_BASIN_TITLE = "hms:basin_title"


class GeojsonAsset(GenericAsset):
    """Geojson asset."""

    regex_parse_str = r".*\.geojson$"
    __roles__ = ["data"]
    __description__ = "Geojson file."
    __media_type__ = MediaType.GEOJSON


class TiffAsset(GenericAsset):
    """Tiff Asset."""

    regex_parse_str = r".*\.tiff$"
    __roles__ = ["data", MediaType.GEOTIFF]
    __description__ = "Tiff file."


class ProjectAsset(GenericAsset[ProjectFile]):
    """Project asset."""

    regex_parse_str = r".*\.hms$"
    __roles__ = ["hms-project", MediaType.TEXT]
    __description__ = "The HEC-HMS project file. Summary provided at the item level"
    __file_class__ = ProjectFile


class ThumbnailAsset(GenericAsset):
    """Thumbnail asset."""

    regex_parse_str = r".*\.png$"
    __roles__ = ["thumbnail", MediaType.PNG]
    __description__ = "Thumbnail"

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        basin_file = self.href.split(".")[0] + ".basin"
        return {"associated_basin_file": basin_file}


class ModelBasinAsset(GenericAsset[BasinFile]):
    """HEC-HMS Basin file asset from authoritative model, containing geometry and other detailed data."""

    regex_parse_str = r".*\.basin$"
    __roles__ = ["hms-basin", MediaType.TEXT]
    __description__ = "Defines the basin geometry and elements for HEC-HMS simulations."
    __file_class__ = BasinFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return (
            {
                HMS_BASIN_TITLE: self.file.name,
                HMS_VERSION: self.file.header.attrs["Version"],
                HMS_DESCRIPTION: self.file.header.attrs.get("Description"),
                HMS_UNIT_SYSTEM: self.file.header.attrs["Unit System"],
                "hms:gages": self.file.gages,
                "hms:drainage_area_miles": self.file.drainage_area,
                "hms:reach_length_miles": self.file.reach_miles,
                "proj:wkt": self.file.wkt,
                "proj:code": self.file.epsg,
            }
            | self.file.hms_methods
            | {f"hms_basin:{key.lower()}": val for key, val in self.file.elements.element_counts.items()}
        )


class EventBasinAsset(GenericAsset[BasinFile]):
    """HEC-HMS Basin file asset from event, with limited basin info."""

    regex_parse_str = r".*\.basin$"
    __roles__ = ["hms-basin", MediaType.TEXT]
    __description__ = "Defines the basin geometry and elements for HEC-HMS simulations."
    __file_class__ = BasinFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {
            HMS_BASIN_TITLE: self.file.name,
            HMS_VERSION: self.file.header.attrs["Version"],
            HMS_DESCRIPTION: self.file.header.attrs.get("Description"),
            HMS_UNIT_SYSTEM: self.file.header.attrs["Unit System"],
        }


class RunAsset(GenericAsset[RunFile]):
    """Run asset."""

    regex_parse_str = r".*\.run$"
    __file_class__ = RunFile
    __roles__ = ["hms-run", MediaType.TEXT]
    __description__ = "Contains data for HEC-HMS simulations."

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {"hms:run_title": self.name.removesuffix(".run")} | {
            run.name: {f"hms:{key}".lower(): val for key, val in run.attrs.items()} for _, run in self.file.elements
        }


class ControlAsset(GenericAsset[ControlFile]):
    """HEC-HMS Control file asset."""

    regex_parse_str = r".*\.control$"
    __roles__ = ["hms-control", MediaType.TEXT]
    __description__ = "Defines time control information for HEC-HMS simulations."
    __file_class__ = ControlFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {
            "hms:control_title": self.file.name,
            **{f"hms:{key}".lower(): val for key, val in self.file.attrs.items()},
        }


class MetAsset(GenericAsset[MetFile]):
    """HEC-HMS Meteorological file asset."""

    regex_parse_str = r".*\.met$"
    __roles__ = ["hms-met", MediaType.TEXT]
    __description__ = "Contains meteorological data such as precipitation and temperature."
    __file_class__ = MetFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {
            "hms:met_title": self.file.name,
            **{f"hms:{key}".lower(): val for key, val in self.file.attrs.items()},
        }


class DSSAsset(GenericAsset):
    """DSS asset."""

    regex_parse_str = r".*\.dss$"
    __roles__ = ["hec-dss", "application/octet-stream"]
    __description__ = "HEC-DSS file."

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.name}


class SqliteAsset(GenericAsset[SqliteDB]):
    """HEC-HMS SQLite database asset."""

    regex_parse_str = r".*\.sqlite$"
    __roles__ = ["hms-sqlite", "application/x-sqlite3"]
    __description__ = "Stores spatial data for HEC-HMS basin files."
    __file_class__ = SqliteDB

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.name, "hms:layers": self.file.layers}


class GageAsset(GenericAsset[GageFile]):
    """Gage asset."""

    regex_parse_str = r".*\.gage$"
    __roles__ = ["hms-gage", MediaType.TEXT]
    __description__ = "Contains data for HEC-HMS gages."
    __file_class__ = GageFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.file.name, HMS_VERSION: self.file.attrs["Version"]} | {
            f"hms:{gage.name}".lower(): dict(gage.attrs.items()) for gage in self.file.gages
        }


class GridAsset(GenericAsset[GridFile]):
    """Grid asset."""

    regex_parse_str = r".*\.grid$"
    __roles__ = ["hms-grid", MediaType.TEXT]
    __description__ = "Contains data for HEC-HMS grid files."
    __file_class__ = GridFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return (
            {HMS_TITLE: self.file.name}
            | {f"hms:{key}".lower(): val for key, val in self.file.attrs.items()}
            | {f"hms:{grid.name}".lower(): dict(grid.attrs.items()) for grid in self.file.grids}
        )


class LogAsset(GenericAsset):
    """Log asset."""

    regex_parse_str = r".*\.log$"
    __roles__ = ["hms-log", "results", MediaType.TEXT]
    __description__ = "Contains log data for HEC-HMS simulations."

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.name}


class OutAsset(GenericAsset):
    """Out asset."""

    regex_parse_str = r".*\.out$"
    __roles__ = ["hms-out", "results", MediaType.TEXT]
    __description__ = "Contains output data for HEC-HMS simulations."

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.name}


class PdataAsset(GenericAsset[PairedDataFile]):
    """Pdata asset."""

    regex_parse_str = r".*\.pdata$"
    __roles__ = ["hms-pdata", MediaType.TEXT]
    __description__ = "Contains paired data for HEC-HMS simulations."
    __file_class__ = PairedDataFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.file.name, HMS_VERSION: self.file.attrs["Version"]}


class TerrainAsset(GenericAsset[TerrainFile]):
    """Terrain asset."""

    regex_parse_str = r".*\.terrain$"
    __roles__ = ["hms-terrain", MediaType.GEOTIFF]
    __description__ = "Contains terrain data for HEC-HMS simulations."
    __file_class__ = TerrainFile

    @GenericAsset.extra_fields.getter
    def extra_fields(self):
        """Return extra fields with added dynamic keys/values."""
        return {HMS_TITLE: self.file.name, HMS_VERSION: self.file.attrs["Version"]} | {
            f"hms:{layer['name']}".lower(): dict(layer.items()) for layer in self.file.layers
        }


HMS_ASSET_CLASSES = [
    ProjectAsset,
    EventBasinAsset,
    ModelBasinAsset,
    ControlAsset,
    MetAsset,
    SqliteAsset,
    GageAsset,
    RunAsset,
    GridAsset,
    LogAsset,
    OutAsset,
    PdataAsset,
    TerrainAsset,
    DSSAsset,
    GeojsonAsset,
    TiffAsset,
    TiffAsset,
    ThumbnailAsset,
]

HMS_EXTENSION_MAPPING = {re.compile(cls.regex_parse_str, re.IGNORECASE): cls for cls in HMS_ASSET_CLASSES}



"""HEC-HMS STAC Item utlity functions."""

from __future__ import annotations

import logging
import re
from collections import OrderedDict

from shapely import MultiPolygon, Polygon

from hecstac.common.logger import get_logger
from hecstac.hms.consts import ATTR_KEYVAL_GROUPER, ATTR_NESTED_KEYVAL_GROUPER, NL_KEYS

logger = get_logger(__name__)


def add_no_duplicate(d: dict, key, val):
    """Insert a key into a dictionary, logging a warning if the key already exists."""
    if key in d:
        return
    d[key] = val


def get_lines_until_end_sentinel(lines: list[str]) -> list[str]:
    """Retrieve all lines until the End point."""
    lines_found = []
    for line in lines:
        if line.strip() in ["End:", "End Computation Point:"]:
            break
        lines_found.append(line)
    else:
        raise ValueError("never found End:")
    return lines_found


def handle_special_cases(key, val):
    """Handle special cases."""
    if key == "Groundwater Layer":
        key = key + val
    elif "Groundwater Layer" in key:
        key = key.rstrip(val)
    elif key == "Use Basin Model":
        key = key + val
    elif "Use Basin Model" in key:
        key = key.rstrip(val)
    return key, val


def parse_attrs(lines: list[str]) -> OrderedDict:
    """Scan the lines down to the first instance of 'End:' and return dict containing all of the colon-separated keyval pair."""
    attrs = {}
    for line in lines:
        if line.strip() == "End:":
            break
        if not line:
            continue

        keyval_pairs = re.findall(ATTR_KEYVAL_GROUPER, line)
        if keyval_pairs:
            key, val = _process_keyval_pairs(attrs, keyval_pairs)

        # test if nested (10 spaces to start the line). if so create nested dictionary
        nested_keyval_pairs = re.findall(ATTR_NESTED_KEYVAL_GROUPER, line)
        if nested_keyval_pairs:
            _process_nested_pair(attrs, nested_keyval_pairs, key, val)

        key, val = _process_hamon_coefficient(attrs, line)
        keyval_pairs = (key, val)

        _handle_keyval_error(line, keyval_pairs, nested_keyval_pairs)
    else:
        raise ValueError("never found End:")
    return OrderedDict(attrs)


def _process_keyval_pairs(attrs: dict, keyval_pairs: list[any]) -> tuple[str, str]:
    """Process a standard key: value pair."""
    if len(keyval_pairs) != 1:
        raise ValueError(f"Expected 0 or 1 pairs, got {len(keyval_pairs)}: {keyval_pairs}")
    key, val = keyval_pairs[0]
    key, val = handle_special_cases(key, val)
    add_no_duplicate(attrs, key, val)

    return (key, val)


def _process_nested_pair(attrs: dict, nested_keyval_pairs: list[any], parent_key: str, parent_val: str):
    """Process a nested key-value pair."""
    if len(nested_keyval_pairs) != 1:
        raise ValueError(f"Expected 0 or 1 pairs, got {len(nested_keyval_pairs)}: {nested_keyval_pairs}")
    nested_key, nested_val = nested_keyval_pairs[0]
    # if attrs[key]val is a string then this is the first item in nested dictionary; change to dictionary instead of string.
    if isinstance(attrs[parent_key], str):
        attrs[parent_key] = {parent_val: {}}
    add_no_duplicate(attrs[parent_key][parent_val], nested_key, nested_val)


def _process_hamon_coefficient(attrs: dict, line: str) -> tuple[str, str]:
    if "Hamon Coefficient" in line:
        key, val = line.split(":")
        add_no_duplicate(attrs, key, val)

    return (key, val)


def _handle_keyval_error(line, keyval_pairs, nested_keyval_pairs):
    if not (keyval_pairs or nested_keyval_pairs):
        raise ValueError(f"unexpected line (does not have keyval nor nested keyval pair): {repr(line)}")
    if keyval_pairs and nested_keyval_pairs:
        raise RuntimeError(f"regex matched top-level and nested keyval pairs (pattern is bugged): {repr(line)}")


def remove_holes(geom):
    """Remove holes in the geometry."""
    if isinstance(geom, Polygon):
        return Polygon(geom.exterior)
    elif isinstance(geom, MultiPolygon):
        areas, polygons = [], []
        for geom in geom.geoms:
            polygons.append(Polygon(geom.exterior))
            areas.append(polygons[-1].area)
        idx = areas.index(max(areas))
        return polygons[idx]
    elif geom is None:
        return geom
    else:
        raise TypeError(f"cannot handle geometry type {type(geom)}. only able to handle Polygon and Multipolygon")


def attrs2list(attrs: OrderedDict) -> list[str]:
    """Convert dictionary of attributes to a list."""
    content = []
    for key, val in attrs.items():
        if isinstance(val, str):
            continue

        if isinstance(val, int) or isinstance(val, float):
            val = str(val)
        elif isinstance(val, dict):
            content += _serialize_dict(key, val)
            continue
        else:
            raise TypeError(
                f"attribute is not of type string. we are currently unable to serialize nested dicts. key, value = {repr(key)}, {repr(val)}"
            )

        content += _format_value(key, val)

    return content


def _serialize_dict(key, val) -> list:
    content = []

    value = list(val.keys())[0]
    content += [f"     {key}: {value}"]
    for k, v in val[value].items():
        content += [f"       {k}: {v}"]

    return content


def _format_value(key, val) -> list:
    content = []

    if key in NL_KEYS:
        content += [""]
    if key in ["Latitude Degrees", "Longitude Degrees"] and len(val.split(".")[0]) == 2:
        val = " " + val

    key, val = handle_special_cases(key, val)
    content += [f"     {key}: {val}"]

    return content


def insert_after_key(dic: dict, insert_key: str, new_key: str, new_val: str) -> OrderedDict:
    """Recreate the dictionary to insert key-val after the occurance of the insert_key if key-val doesn't exist yet in the dictionary."""
    new_dic = {}
    for key, val in dic.items():
        if key == new_key:
            continue
        new_dic[key] = val
        if key == insert_key:
            new_dic[new_key] = new_val
    return OrderedDict(new_dic)


def search_contents(lines: list, search_string: str, token: str = "=", expect_one: bool = True) -> list[str]:
    """Split a line by a token and returns the second half of the line if the search_string is found in the first half."""
    results = []
    for line in lines:
        if f"{search_string}{token}" in line:
            results.append(line.split(token)[1])

    if expect_one and len(results) > 1:
        raise ValueError(f"expected 1 result, got {len(results)}")
    elif expect_one and len(results) == 0:
        raise ValueError("expected 1 result, no results found")
    elif expect_one and len(results) == 1:
        return results[0]
    else:
        return results


class StacPathManager:
    """Build consistent paths for STAC items and collections assuming a top level local catalog."""

    def __init__(self, local_catalog_dir: str):
        self._catalog_dir = local_catalog_dir

    @property
    def catalog_dir(self):
        """Return the catalog directory."""
        return self._catalog_dir

    @property
    def catalog_file(self):
        """Return the catalog file path."""
        return f"{self._catalog_dir}/catalog.json"

    def catalog_item(self, item_id: str) -> str:
        """Return the catalog item file path."""
        return f"{self.catalog_dir}/{item_id}/{item_id}.json"

    def catalog_asset(self, item_id: str, asset_dir: str = "hydro_domains") -> str:
        """Return the catalog asset file path."""
        return f"{self.catalog_dir}/{asset_dir}/{item_id}.json"

    def collection_file(self, collection_id: str) -> str:
        """Return the collection file path."""
        return f"{self.catalog_dir}/{collection_id}/collection.json"

    def collection_dir(self, collection_id: str) -> str:
        """Return the collection directory."""
        return f"{self.catalog_dir}/{collection_id}"

    def collection_asset(self, collection_id: str, filename: str) -> str:
        """Return the collection asset filepath."""
        return f"{self.catalog_dir}/{collection_id}/{filename}"

    def collection_item_dir(self, collection_id: str, item_id: str) -> str:
        """Return the collection item directory."""
        return f"{self.catalog_dir}/{collection_id}/{item_id}"

    def collection_item(self, collection_id: str, item_id: str) -> str:
        """Return the collection item filepath."""
        return f"{self.catalog_dir}/{collection_id}/{item_id}/{item_id}.json"

    def collection_item_asset(self, collection_id: str, item_id: str, filename: str) -> str:
        """Return the collection item asset filepath."""
        return f"{self.catalog_dir}/{collection_id}/{item_id}/{filename}"



"""Logging utility and setup."""

import json
import logging
import sys
from re import L

SUPPRESS_LOGS = [
    "boto3",
    "botocore",
    "geopandas",
    "fiona",
    "rasterio",
    "pyogrio",
    "xarray",
    "shapely",
    "matplotlib",
    "aiobotocore",
    "fsspec",
]


class JsonFormatter(logging.Formatter):
    """Custom JSON formatter."""

    def format(self, record):
        """Add name and format structure."""
        logger_name = record.name
        if logger_name == "root":
            logger_name = "hecstac"

        log_record = {
            "time": self.formatTime(record, datefmt="%Y-%m-%dT%H:%M:%SZ"),
            "level": record.levelname,
            "logger": logger_name,
            "message": record.getMessage(),
        }
        return json.dumps(log_record)


def initialize_logger(json_logging: bool = False, level: int = logging.INFO) -> logging.Logger:
    """Initialize and return the hecstac logger."""
    logger = logging.getLogger("hecstac")
    logger.setLevel(level)

    if not logger.handlers:
        if json_logging:

            class FlushStreamHandler(logging.StreamHandler):
                def emit(self, record):
                    super().emit(record)
                    self.flush()

            handler = FlushStreamHandler(sys.stdout)
            formatter = JsonFormatter()
            handler.setFormatter(formatter)
        else:
            handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter("%(asctime)s %(levelname)s %(name)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
            handler.setFormatter(formatter)

        handler.setLevel(level)
        logger.addHandler(handler)
        logger.propagate = False  # don't double log to root

        # Configure the root logger
        root_logger = logging.getLogger()
        root_logger.setLevel(level)
        if not root_logger.handlers:
            root_logger.addHandler(handler)

    for package in SUPPRESS_LOGS:
        logging.getLogger(package).setLevel(logging.WARNING)

    return logger


def get_logger(name: str = None) -> logging.Logger:
    """Get hecstac logger or a child logger."""
    base_logger = logging.getLogger("hecstac")
    if name:
        return base_logger.getChild(name)
    return base_logger



"""HEC-HMS file parsing classes."""

from __future__ import annotations

import logging
import math
import os
from abc import ABC
from collections import OrderedDict
from functools import lru_cache
from pathlib import Path

import fiona
import geopandas as gpd
import pandas as pd
from pyparsing import Optional
from pyproj import CRS
from shapely.geometry import LineString, MultiLineString, Point

import hecstac.hms.utils as utils
from hecstac.common.base_io import ModelFileReader
from hecstac.common.consts import S3_PREFIX
from hecstac.common.logger import get_logger
from hecstac.hms.consts import BC_LENGTH, BC_LINE_BUFFER
from hecstac.hms.data_model import (
    ET,
    BasinHeader,
    BasinLayerProperties,
    BasinSchematicProperties,
    BasinSpatialProperties,
    ComputationPoints,
    Diversion,
    ElementSet,
    Gage,
    Grid,
    Junction,
    Pattern,
    Precipitation,
    Reach,
    Reservoir,
    Run,
    Sink,
    Source,
    Subbasin,
    SubbasinET,
    Table,
    Temperature,
)
from hecstac.hms.s3_utils import create_fiona_aws_session

logger = get_logger(__name__)

HEADER_BASIN = "Basin: "
HEADER_SUBBASIN = "Subbasin: "
HEADER_FILENAME = "     Filename: "
HEADER_CONTROL = "Control: "
HEADER_TERRAIN_DATA = "Terrain Data: "
OBSERVED_HYDROGRAPH_GAGE = "Observed Hydrograph Gage"
CANVAS_X = "Canvas X"
CANVAS_Y = "Canvas Y"
DSS_FILE = "DSS File"


class BaseTextFile(ABC):
    """Base class for text files."""

    def __init__(self, path: str):
        self.path: str = path
        self.directory: str = os.path.dirname(self.path)
        self.stem: str = os.path.splitext(self.path)[0]
        self.content: Optional[str] = None
        self.attrs: dict = {}
        self.read_content()
        self.parse_header()

    def read_content(self):
        """Read contents of text file."""
        if os.path.exists(self.path):
            try:
                with open(self.path, encoding="utf-8") as f:
                    self.content = f.read()
            except UnicodeDecodeError:
                with open(self.path, encoding="cp1252") as f:
                    self.content = f.read()
        else:
            try:
                self.model_file = ModelFileReader(self.path)
                self.content = self.model_file.content
            except Exception as e:
                logger.error(e)
                raise FileNotFoundError(f"could not find {self.path} locally nor on s3")

    def parse_header(self):
        """Scan the file down to the first instance of 'End:' and save each colon-separated keyval pair as attrs dict."""
        lines = self.content.splitlines()
        if not lines[0].startswith(
            (
                "Project:",
                "Basin:",
                "Meteorology:",
                "Control:",
                "Terrain Data Manager:",
                "Run:",
                "Paired Data Manager:",
                "Grid Manager",
                "Gage Manager",
            )
        ):
            raise ValueError(f"Unexpected first line: {lines[0]}")
        self.attrs = utils.parse_attrs(lines[1:])


class ProjectFile(BaseTextFile):
    """Class for parsing HEC-HMS project files."""

    def __init__(
        self,
        path: str,
        recurse: bool = True,
        assert_uniform_version: bool = True,
    ):
        if not path.endswith(".hms"):
            raise ValueError(f"invalid extension for Project file: {path}")
        super().__init__(path)

        self.basins = []
        self.mets = []
        self.controls = []
        self.terrain = None
        self.run = None
        self.gage = None
        self.grid = None
        self.pdata = None
        self.logger = get_logger(__name__)

        if recurse:
            self.scan_for_basins_mets_controls()
            self.scan_for_terrain_run_grid_gage_pdata()
            if assert_uniform_version:
                self.assert_uniform_version()

    def __repr__(self):
        """Representation of the HMSProjectFile class."""
        return f"HMSProjectFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Extract name from project file."""
        lines = self.content.splitlines()
        if not lines[0].startswith("Project: "):
            raise ValueError(f"unexpected first line: {lines[0]}")
        return lines[0][len("Project: ") :]

    def combine_stem_ext(self, ext: str) -> str:
        """Combine stem and extension."""
        return f"{self.stem}.{ext}"

    def scan_for_terrain_run_grid_gage_pdata(self):
        """Scan for terrain, run, grid, gage, and pdata files."""
        for ext in ["terrain", "run", "grid", "gage", "pdata"]:
            path = self.combine_stem_ext(ext)
            if os.path.exists(path):
                if ext == "terrain":
                    self.terrain = TerrainFile(path)
                elif ext == "run":
                    self.run = RunFile(path)
                elif ext == "grid":
                    self.grid = GridFile(path)
                elif ext == "gage":
                    self.gage = GageFile(path)
                elif ext == "pdata":
                    self.pdata = PairedDataFile(path)

    def scan_for_basins_mets_controls(self):
        """Scan for basin, meteorology, and control files."""
        lines = self.content.splitlines()
        i = -1
        while True:
            i += 1
            if i >= len(lines):
                break
            line = lines[i]

            if line.startswith(HEADER_BASIN):
                self._find_basin_file_path(lines, i)

            if line.startswith("Precipitation: "):
                self._find_precipitation_file_path(lines, i)

            if line.startswith(HEADER_CONTROL):
                self._find_control_file_path(lines, i)

    def _find_basin_file_path(self, lines, i):
        nextline = lines[i + 1]
        if not nextline.startswith(HEADER_FILENAME):
            raise ValueError(f"unexpected line: {nextline}")
        basinfile_bn = nextline[len(HEADER_FILENAME) :]
        if self.directory.startswith(S3_PREFIX):
            basinfile_path = f"{self.directory}/{basinfile_bn}"
        else:
            basinfile_path = os.path.join(self.directory, basinfile_bn)
        self.basins.append(BasinFile(basinfile_path))

    def _find_precipitation_file_path(self, lines, i):
        nextline = lines[i + 1]
        if not nextline.startswith(HEADER_FILENAME):
            raise ValueError(f"unexpected line: {nextline}")
        metfile_bn = nextline[len(HEADER_FILENAME) :]
        if self.directory.startswith(S3_PREFIX):
            metfile_path = f"{self.directory}/{metfile_bn}"
        else:
            metfile_path = os.path.join(self.directory, metfile_bn)
        self.mets.append(MetFile(metfile_path))

    def _find_control_file_path(self, lines, i):
        nextline = lines[i + 1]
        if not nextline.startswith("     FileName: "):
            raise ValueError(f"unexpected line: {nextline}")
        controlfile_bn = nextline[len("     FileName: ") :]
        if self.directory.startswith(S3_PREFIX):
            controlfile_path = f"{self.directory}/{controlfile_bn}"
        else:
            controlfile_path = os.path.join(self.directory, controlfile_bn)
        self.controls.append(ControlFile(controlfile_path))

    @property
    def file_counts(self):
        """Return file counts."""
        return {
            "Basins": len(self.basins),
            "Controls": len(self.controls),
            "Mets": len(self.mets),
            "Runs": 1 if self.run is not None else None,
            "Terrain": 1 if self.terrain is not None else None,
            "Paired_Data": 1 if self.pdata is not None else None,
            "Grid": 1 if self.grid is not None else None,
            "Gage": 1 if self.gage is not None else None,
            "SQLite": len([basin.sqlite_path for basin in self.basins if os.path.exists(basin.sqlite_path)]),
        }

    def assert_uniform_version(self):
        """Assert uniform version."""
        errors = []
        version = self.attrs["Version"]
        for basin in self.basins:
            if basin.attrs["Version"] != version:
                errors.append(f"Basin {basin.name} version mismatch (expected {version}, got {basin.attrs['Version']})")
        for met in self.mets:
            if met.attrs["Version"] != version:
                errors.append(
                    f"Meteorology {met.name} version mismatch (expected {version}, got {met.attrs['Version']})"
                )
        for control in self.controls:
            if control.attrs["Version"] != version:
                errors.append(
                    f"Control {control.name} version mismatch (expected {version}, got {control.attrs['Version']})"
                )
        if self.terrain and self.terrain.attrs["Version"] != version:
            errors.append(f"Terrain version mismatch (expected {version}, got {self.terrain.attrs['Version']})")
        # RunFile has no version
        if errors:
            raise ValueError("\n".join(errors))

    @property
    def files(self):
        """Return associated files."""
        # self.logger.info(f"other paths {[i.path for i in [self.terrain, self.run, self.grid, self.gage, self.pdata] if i]}")
        return (
            [self.path]
            + [basin.path for basin in self.basins]
            + [basin.sqlite_path for basin in self.basins]
            + [control.path for control in self.controls]
            + [met.path for met in self.mets]
            + [i.path for i in [self.terrain, self.run, self.grid, self.gage, self.pdata] if i]
            + self.result_files
            + self.dss_files
        )

    @property
    def dss_files(self):
        """Return dss files."""
        files = set()
        if self.gage:
            try:
                files.update(
                    [
                        gage.attrs["Variant"]["Variant-1"]["DSS File Name"]
                        for gage in self.gage.elements.elements.values()
                    ]
                )
            except KeyError:
                files.update([gage.attrs["Filename"] for gage in self.gage.elements.elements.values()])
        else:
            logger.warning("No gage file to extract gages from.")

        if self.pdata:
            files.update([pdata.attrs[DSS_FILE] for pdata in self.pdata.elements.elements.values()])
        else:
            logger.warning("No pdata files found.")

        if self.grid:
            files.update(
                [
                    grid.attrs["Variant"]["Variant-1"]["DSS File Name"]
                    for grid in self.grid.elements.elements.values()
                    if "Variant" in grid.attrs
                ]
            )
        else:
            logger.warning("No grid file to extract dss files from.")

        files = [str(Path(f.replace("\\", "/"))) for f in files]
        return self.absolute_paths(files)

    @property
    def result_files(self):
        """Return result files."""
        if self.run:
            files = set(
                [i[1].attrs["Log File"] for i in self.run.elements]
                + [i[1].attrs[DSS_FILE] for i in self.run.elements]
                + [i[1].attrs[DSS_FILE].replace(".dss", ".out") for i in self.run.elements]
            )

            files = [str(Path(f.replace("\\", "/"))) for f in files]
            return self.absolute_paths(set(files))
        else:
            return []

    def absolute_paths(self, paths):
        """Return absolute path."""
        return [os.path.join(self.directory, path) for path in paths]

    @property
    def rasters(self):
        """Return raster files."""
        files = []

        if self.terrain:
            for terrain in self.terrain.layers:
                raster_dir = terrain.get("raster_dir", "").strip()
                if raster_dir and os.path.exists(raster_dir):
                    files += [os.path.join(raster_dir, f) for f in os.listdir(raster_dir)]
                else:
                    logger.warning(f"Skipping missing raster directory: {raster_dir}")

        if self.grid is None:
            logger.warning("No grid file, skipping grid rasters.")
        else:
            files += [
                grid.attrs["Filename"] for grid in self.grid.elements.elements.values() if "Filename" in grid.attrs
            ]
        files = [str(Path(f.replace("\\", "/"))) for f in files]
        return self.absolute_paths(set(files))

    @property
    @lru_cache
    def sqlitedbs(self):
        """Return SQLite database."""
        return [SqliteDB(basin.sqlite_path) for basin in self.basins]


class BasinFile(BaseTextFile):
    """Class for parsing HEC-HMS basin files."""

    def __init__(
        self,
        path: str,
        skip_scans: bool = False,
        fiona_aws_session=None,
        read_geom: bool = True,
    ):
        if not path.endswith(".basin"):
            raise ValueError(f"invalid extension for Basin file: {path}")
        super().__init__(path)

        self.header: Optional[BasinHeader] = None
        self.layer_properties: Optional[BasinLayerProperties] = None
        self.spatial_properties: Optional[BasinSpatialProperties] = None
        self.schematic_properties: Optional[BasinSchematicProperties] = None
        self.computation_points: Optional[ComputationPoints] = None
        self.fiona_aws_session = fiona_aws_session
        self.read_geom = read_geom
        self.name = None
        self.parse_name()

        if not skip_scans:
            self.scan_for_headers_and_footers()

        if self.read_geom:
            sqlite_basename = self.identify_sqlite()

            if self.path.startswith(S3_PREFIX):
                self.sqlite_path = f"{os.path.dirname(self.path)}/{sqlite_basename}"
                self.fiona_aws_session = create_fiona_aws_session()
            else:
                self.sqlite_path = os.path.join(os.path.dirname(self.path), sqlite_basename)

    def __repr__(self):
        """Representation of the HMSBasinFile class."""
        return f"HMSBasinFile({self.path})"

    @property
    def wkt(self):
        """Return wkt representation of the CRS."""
        for line in self.spatial_properties.content.splitlines():
            if "Coordinate System: " in line:
                return line.split(": ")[1]

    @property
    def crs(self):
        """Return the CRS."""
        return CRS(self.wkt)

    @property
    def epsg(self):
        """Return the EPSG code."""
        return self.crs.to_epsg()

    def parse_name(self):
        """Parse basin name."""
        lines = self.content.splitlines()
        if not lines[0].startswith(HEADER_BASIN):
            raise ValueError(f"unexpected first line: {lines[0]}")
        self.name = lines[0][len(HEADER_BASIN) :]

    def scan_for_headers_and_footers(self):
        """Scan for basin headers and footers."""
        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith(HEADER_BASIN):
                attrs = utils.parse_attrs(lines[i + 1 :])
                self.header = BasinHeader(attrs)
            if line.startswith("Basin Schematic Properties:"):
                attrs = utils.parse_attrs(lines[i + 1 :])
                self.schematic_properties = BasinSchematicProperties(attrs)
            if line.startswith("Basin Spatial Properties:"):
                content = "\n".join(utils.get_lines_until_end_sentinel(lines[i + 1 :]))
                self.spatial_properties = BasinSpatialProperties(content)
            if line.startswith("Basin Layer Properties:"):
                content = "\n".join(utils.get_lines_until_end_sentinel(lines[i + 1 :]))
                self.layer_properties = BasinLayerProperties(content)
            if line.startswith("Computation Point:"):
                content = "\n".join(utils.get_lines_until_end_sentinel(lines[i + 1 :]))
                self.computation_points = ComputationPoints(content)

    def identify_sqlite(self):
        """Identify SQLite."""
        for line in self.content.splitlines():
            if ".sqlite" in line:
                return line.split("File: ")[1]

    @property
    @lru_cache
    def elements(self):
        """Return basin elements."""
        elements = ElementSet()
        if self.read_geom:
            sqlite = SqliteDB(
                self.sqlite_path,
                fiona_aws_session=self.fiona_aws_session,
            )

        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith(HEADER_SUBBASIN):
                name = line[len(HEADER_SUBBASIN) :]
                elements[name] = self._parse_subbasin(lines, name, sqlite, i)

            elif line.startswith("Reach: "):
                name = line[len("Reach: ") :]
                elements[name] = self._parse_reach(lines, name, sqlite, i)

            elif line.startswith("Junction: "):
                name = line[len("Junction: ") :]
                elements[name] = self._parse_junction(lines, name, i)

            elif line.startswith("Sink: "):
                name = line[len("Sink: ") :]
                elements[name] = self._parse_sink(lines, name, i)

            elif line.startswith("Reservoir: "):
                name = line[len("Reservoir: ") :]
                elements[name] = self._parse_reservoir(lines, name, i)

            elif line.startswith("Source: "):
                name = line[len("Source: ") :]
                elements[name] = self._parse_source(lines, name, i)

            elif line.startswith("Diversion: "):
                name = line[len("Diversion: ") :]
                elements[name] = self._parse_diversion(lines, name, i)

        return elements

    def _parse_subbasin(self, lines, name, sqlite, i) -> Subbasin:
        geom = None
        attrs = utils.parse_attrs(lines[i + 1 :])
        if self.read_geom:
            geom = sqlite.subbasin_feats[sqlite.subbasin_feats["name"] == name].geometry.values[0]

        return Subbasin(name, attrs, geom)

    def _parse_reach(self, lines, name, sqlite, i) -> Reach:
        geom = None
        slope = None
        attrs = utils.parse_attrs(lines[i + 1 :])
        if self.read_geom:
            try:
                geom = sqlite.reach_feats[sqlite.reach_feats["name"] == name].geometry.values[0]
                if "slope" in sqlite.reach_feats.columns:
                    slope = sqlite.reach_feats[sqlite.reach_feats["name"] == name]["slope"].values[0]
                else:
                    slope = 0
            except IndexError:
                x1 = utils.search_contents(lines[i + 1 :], CANVAS_X, ":", False)[0]
                y1 = utils.search_contents(lines[i + 1 :], CANVAS_Y, ":", False)[0]
                x2 = utils.search_contents(lines[i + 1 :], f"From {CANVAS_X}", ":", False)[0]
                y2 = utils.search_contents(lines[i + 1 :], f"From {CANVAS_Y}", ":", False)[0]
                geom = LineString([[float(x1), float(y1)], [float(x2), float(y2)]])
                slope = 0

        return Reach(name, attrs, geom, slope)

    def _parse_junction(self, lines, name, i) -> Junction:
        geom = None
        attrs = utils.parse_attrs(lines[i + 1 :])
        if self.read_geom:
            geom = Point((float(attrs[CANVAS_X]), float(attrs[CANVAS_Y])))

        return Junction(name, attrs, geom)

    def _parse_sink(self, lines, name, i) -> Sink:
        geom = None
        attrs = utils.parse_attrs(lines[i + 1 :])
        if self.read_geom:
            geom = Point((float(attrs[CANVAS_X]), float(attrs[CANVAS_Y])))

        return Sink(name, attrs, geom)

    def _parse_reservoir(self, lines, name, i) -> Reservoir:
        geom = None
        attrs = OrderedDict({"text": lines[i + 1 :]})
        if self.read_geom:
            x = utils.search_contents(lines[i + 1 :], CANVAS_X, ":", False)[0]
            y = utils.search_contents(lines[i + 1 :], CANVAS_Y, ":", False)[0]
            geom = Point((float(x), float(y)))

        return Reservoir(name, attrs, geom)

    def _parse_source(self, lines, name, i) -> Source:
        geom = None
        attrs = utils.parse_attrs(lines[i + 1 :])
        if self.read_geom:
            geom = Point((float(attrs[CANVAS_X]), float(attrs[CANVAS_Y])))

        return Source(name, attrs, geom)

    def _parse_diversion(self, lines, name, i) -> Diversion:
        geom = None
        attrs = utils.parse_attrs(lines[i + 1 :])
        if self.read_geom:
            geom = Point((float(attrs[CANVAS_X]), float(attrs[CANVAS_Y])))

        return Diversion(name, attrs, geom)

    @property
    @lru_cache
    def subbasins(self):
        """Return subbasin elements."""
        return self.elements.get_element_type("Subbasin")

    @property
    @lru_cache
    def reaches(self):
        """Return reach elements."""
        return self.elements.get_element_type("Reach")

    @property
    @lru_cache
    def junctions(self):
        """Return junction elements."""
        return self.elements.get_element_type("Junction")

    @property
    @lru_cache
    def reservoirs(self):
        """Return reservoir elements."""
        return self.elements.get_element_type("Reservoir")

    @property
    @lru_cache
    def diversions(self):
        """Return diversion elements."""
        return self.elements.get_element_type("Diversion")

    @property
    @lru_cache
    def sinks(self):
        """Return sink elements."""
        return self.elements.get_element_type("Sink")

    @property
    @lru_cache
    def sources(self):
        """Return source elements."""
        return self.elements.get_element_type("Source")

    @property
    @lru_cache
    def gages(self):
        """Return gages."""
        return self.elements.gages

    @property
    @lru_cache
    def drainage_area(self):
        """Return drainage areas.."""
        return sum([subbasin.geom.area for subbasin in self.subbasins])

    @property
    @lru_cache
    def reach_miles(self):
        """Return reach lengths in miles.."""
        return sum([reach.geom.length for reach in self.reaches])

    @property
    @lru_cache
    def basin_geom(self):
        """Return basin geometry."""
        return utils.remove_holes(self.feature_2_gdf("Subbasin").make_valid().to_crs(4326).union_all())

    def bbox(self, crs):
        """Return basin bounding box."""
        return self.feature_2_gdf("Subbasin").to_crs(crs).total_bounds

    def feature_2_gdf(self, element_type: str) -> gpd.GeoDataFrame:
        """Convert feature to GeoDataFrame."""
        gdf_list = []
        for e in self.elements.get_element_type(element_type):
            gdf_list.append(
                gpd.GeoDataFrame([{"name": e.name, "geometry": e.geom} | e.attrs], geometry="geometry", crs=self.crs)
            )
        if len(gdf_list) == 1:
            return gdf_list[0]
        elif len(gdf_list) == 0:
            return None
        else:
            return pd.concat(gdf_list)

    @property
    @lru_cache
    def observation_points_gdf(self):
        """Return GeoDataFrame of observation points."""
        gdf_list = []
        for name, element in self.elements:
            if OBSERVED_HYDROGRAPH_GAGE in element.attrs.keys():
                if isinstance(element, Junction) or isinstance(element, Sink):
                    gdf_list.append(
                        gpd.GeoDataFrame(
                            {
                                "name": name,
                                "geometry": element.geom,
                                "gage_name": element.attrs[OBSERVED_HYDROGRAPH_GAGE],
                            },
                            geometry="geometry",
                            crs=self.crs,
                            index=[0],
                        )
                    )
                elif isinstance(element, Subbasin):
                    gdf_list.append(
                        gpd.GeoDataFrame(
                            {
                                "name": name,
                                "geometry": element.geom.centroid,
                                "gage_name": element.attrs[OBSERVED_HYDROGRAPH_GAGE],
                            },
                            geometry="geometry",
                            crs=self.crs,
                            index=[0],
                        )
                    )
                elif isinstance(element, Reach):
                    start_point = element.geom.boundary
                    gdf_list.append(
                        gpd.GeoDataFrame(
                            {
                                "name": name,
                                "geometry": start_point,
                                "gage_name": element.attrs[OBSERVED_HYDROGRAPH_GAGE],
                            },
                            geometry="geometry",
                            crs=self.crs,
                            index=[0],
                        )
                    )
        if len(gdf_list) == 1:
            return gdf_list[0]
        elif len(gdf_list) == 0:
            return None
        else:
            gdf = gpd.GeoDataFrame(pd.concat(gdf_list), crs=self.crs, geometry="geometry")
            return gdf

    def subbasin_connection_lines(self) -> gpd.GeoDataFrame:
        """Return GeoDataframe of subbasin connection lines."""
        df_list = []
        for subbasin in self.subbasins:
            us_point = subbasin.geom.centroid
            ds_element = self.elements[subbasin.attrs["Downstream"]]
            if ds_element in self.reaches:
                ds_point = Point(ds_element.geom.coords[-1])
            else:
                ds_point = ds_element.geom
            df = pd.DataFrame(subbasin.attrs, index=[0])
            if not us_point.equals(ds_point):
                df["us_name"], df["ds_name"], df["geometry"] = (
                    subbasin.name,
                    ds_element.name,
                    LineString([us_point, ds_point]),
                )
            df_list.append(df)
        df = pd.concat(df_list)
        gdf = gpd.GeoDataFrame(df, geometry="geometry", crs=self.crs)
        return gdf

    def junction_connection_lines(self) -> gpd.GeoDataFrame:
        """Return GeoDataframe of junction connection lines."""
        df_list = []
        for junction in self.junctions:
            us_point = junction.geom
            if "Downstream" not in junction.attrs:
                logger.warning(f"Warning no downstream element for junction {junction.name}")
                continue
            ds_element = self.elements[junction.attrs["Downstream"]]
            if ds_element in self.reaches:
                if isinstance(ds_element.geom, LineString):
                    ds_point = Point(ds_element.geom.coords[-1])
                elif isinstance(ds_element.geom, MultiLineString):
                    ds_point = Point(ds_element.geom.geoms[0].coords[-1])
                else:
                    raise TypeError(
                        f"Expected either LineString or MultiLineString for reaches; recieved {type(ds_element.geom)}"
                    )
            else:
                ds_point = ds_element.geom
            df = pd.DataFrame(junction.attrs, index=[0])
            df["us_point"] = us_point
            if not us_point.equals(ds_point):
                df["us_name"], df["ds_name"], df["geometry"] = (
                    junction.name,
                    ds_element.name,
                    LineString([Point(us_point.x, us_point.y), Point(ds_point.x, ds_point.y)]),
                )
            df_list.append(df)
        df = pd.concat(df_list)
        if "geometry" in df.columns:
            df = df.drop(columns=["us_point"])
            gdf = gpd.GeoDataFrame(df, geometry="geometry", crs=self.crs)
        else:
            gdf = gpd.GeoDataFrame(df, geometry="us_point", crs=self.crs)
        return gdf

    @property
    @lru_cache
    def hms_schematic_2_gdfs(self) -> dict[gpd.GeoDataFrame]:
        """Convert HMS schematics to GeoDataframe."""
        element_gdfs = {}
        for element_type in [
            "Reach",
            "Subbasin",
            "Junction",
            "Diversion",
            "Source",
            "Sink",
            "Reservoir",
        ]:
            if self.elements.get_element_type(element_type):
                element_gdfs[element_type] = self.feature_2_gdf(element_type)
        element_gdfs["Subbasin_Connectors"] = self.subbasin_connection_lines()
        element_gdfs["Junction_Connectors"] = self.junction_connection_lines()
        element_gdfs["Recommended_BC_Lines"] = self.subbasin_bc_lines()
        return element_gdfs

    def subbasin_bc_lines(self):
        """Return subbasin boundary condition lines."""
        df_list = []
        for _, row in self.subbasin_connection_lines().iterrows():
            geom = row.geometry
            p1 = Point(geom.coords[0])
            p2 = Point(geom.coords[1])
            p3 = geom.interpolate(geom.length - BC_LINE_BUFFER)
            reach_angle = math.atan2(p2.y - p1.y, p2.x - p1.x)  # atan2(y1 - y0, x1 - x0)
            # rotate cross section direction to be quarter of a turn clockwise from reach direction
            bc_angle = reach_angle - math.pi / 2
            x_start = p3.x - math.cos(bc_angle) * BC_LENGTH / 2
            y_start = p3.y - math.sin(bc_angle) * BC_LENGTH / 2
            x_end = p3.x + math.cos(bc_angle) * BC_LENGTH / 2
            y_end = p3.y + math.sin(bc_angle) * BC_LENGTH / 2
            bc_geom = LineString(((x_start, y_start), (x_end, y_end)))
            df_list.append(pd.DataFrame([[row["us_name"], bc_geom]], columns=["name", "geometry"]))
        df = pd.concat(df_list)
        gdf = gpd.GeoDataFrame(df, geometry="geometry", crs=self.crs)
        return gdf

    @property
    @lru_cache
    def hms_methods(self):
        """Extract unique HMS methods from Subbasins and Reaches."""
        methods = {
            "Canopy": set(),
            "Discretization": set(),
            "Begin Snow": set(),
            "Surface": set(),
            "LossRate": set(),
            "Transform": set(),
            "Baseflow": set(),
            "Route": set(),
        }

        for subbasin in self.subbasins:
            for key in methods.keys():
                if key in subbasin.attrs:
                    methods[key].add(subbasin.attrs[key])
        for reach in self.reaches:
            if "Route" in reach.attrs:
                methods["Route"].add(reach.attrs["Route"])
        return {f"hms_methods:{key.replace(' ', '_')}": list(values) for key, values in methods.items()}


class MetFile(BaseTextFile):
    """Class for parsing HEC-HMS meteorology files."""

    def __init__(self, path: str):
        if not path.endswith(".met"):
            raise ValueError(f"invalid extension for Meteorology file: {path}")
        super().__init__(path)

        self.scan_for_elements()

    def __repr__(self):
        """Representation of the HMSMetFile class."""
        return f"HMSMetFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Return meteorology name."""
        lines = self.content.splitlines()
        if not lines[0].startswith("Meteorology: "):
            raise ValueError(f"unexpected first line: {lines[0]}")
        return lines[0][len("Meteorology: ") :]

    def scan_for_elements(self):
        """Scan for meteorology elements."""
        elements = ElementSet()
        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith("Precip Method Parameters: "):
                name = line[len("Precip Method Parameters: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                elements[name] = Precipitation(name=name, attrs=attrs)

            elif line.startswith("Air Temperature Method Parameters: "):
                name = line[len("Air Temperature Method Parameters: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                elements[name] = Temperature(name=name, attrs=attrs)

            elif line.startswith("Evapotranspiration Method Parameters: "):
                name = line[len("Evapotranspiration Method Parameters: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                elements[name] = ET(name=name, attrs=attrs)

            elif line.startswith(HEADER_SUBBASIN):
                name = line[len(HEADER_SUBBASIN) :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                elements[name] = SubbasinET(name=name, attrs=attrs)
        self.elements = elements


class ControlFile(BaseTextFile):
    """Class for parsing HEC-HMS control files."""

    def __init__(self, path: str):
        if not path.endswith(".control"):
            raise ValueError(f"invalid extension for Control file: {path}")
        super().__init__(path)

    def __repr__(self):
        """Representation of the HMSControlFile class."""
        return f"HMSControlFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Return control name."""
        lines = self.content.splitlines()
        if not lines[0].startswith(HEADER_CONTROL):
            raise ValueError(f"unexpected first line: {lines[0]}")
        return lines[0][len(HEADER_CONTROL) :]


class TerrainFile(BaseTextFile):
    """Class for parsing HEC-HMS terrain files."""

    def __init__(self, path: str):
        if not path.endswith(".terrain"):
            raise ValueError(f"Invalid extension for Terrain file: {path}")
        super().__init__(path)
        self.layers = []

        found_first = False
        name, raster_path, raster_dir, vert_units = "", "", "", ""

        for line in self.content.splitlines():
            if not found_first:
                if line.startswith(HEADER_TERRAIN_DATA):
                    found_first = True
                else:
                    continue

            if line.strip() == "End:":
                self.layers.append(
                    {
                        "name": name,
                        "raster_path": raster_path,
                        "raster_dir": raster_dir,
                        "vert_units": vert_units,
                    }
                )
                name, raster_path, raster_dir, vert_units = "", "", "", ""

            elif line.startswith(HEADER_TERRAIN_DATA):
                name = line[len(HEADER_TERRAIN_DATA) :]
            elif line.startswith("     Elevation File Name: "):
                raster_path_raw = line[len("     Elevation File Name: ") :]
                raster_path = os.path.join(os.path.dirname(self.path), raster_path_raw.replace("\\", os.sep))
            elif line.startswith("     Terrain Directory: "):
                raster_dir_raw = line[len("     Terrain Directory: ") :]
                raster_dir = os.path.join(os.path.dirname(self.path), raster_dir_raw.replace("\\", os.sep))
            elif line.startswith("     Vertical Units: "):
                vert_units = line[len("     Vertical Units: ") :]

    def __repr__(self):
        """Representation of the HMSTerrainFile class."""
        return f"HMSTerrainFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Return name."""
        return None


class RunFile(BaseTextFile):
    """Class for parsing HEC-HMS run files."""

    def __init__(self, path: str):
        if not path.endswith(".run"):
            raise ValueError(f"invalid extension for Run file: {path}")
        super().__init__(path)

    def __repr__(self):
        """Representation of the HMSRunFile class."""
        return f"HMSRunFile({self.path})"

    def runs(self):
        """Retrieve all runs."""
        runs = ElementSet()
        lines = self.content.splitlines()
        i = -1
        while True:
            i += 1
            if i >= len(lines):
                break
            line = lines[i]
            if line.startswith("Run: "):
                name = line.split("Run: ")[1]
                runs[name] = Run(name, utils.parse_attrs(lines[i + 1 :]))
        return runs

    @property
    def elements(self):
        """Return run elements."""
        return self.runs()


class PairedDataFile(BaseTextFile):
    """Class for parsing HEC-HMS paired data files."""

    def __init__(self, path: str, client=None, bucket=None):
        if not path.endswith(".pdata"):
            raise ValueError(f"invalid extension for Paired Data file: {path}")
        super().__init__(path)
        self.elements = ElementSet()
        self.scan_for_tables()

    def __repr__(self):
        """Representation of the HMSPairedDataFile class."""
        return f"HMSPairedDataFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Return paired data manager."""
        lines = self.content.splitlines()
        if not lines[0].startswith("Paired Data Manager: "):
            raise ValueError(f"unexpected first line: {lines[0]}")
        return lines[0][len("Paired Data Manager: ") :]

    def scan_for_tables(self):
        """Scan for tables."""
        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith("Table: "):
                name = line[len("Table: ") :]
                table_type = lines[i + 1][len("     Table Type: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                self.elements[f"{name}+{table_type}"] = Table(name, attrs)

    def scan_for_patterns(self):
        """Scan for patterns."""
        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith("Pattern: "):
                name = line[len("Pattern: ") :]
                data_type = lines[i + 1][len("     Data Type: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                self.elements[f"{name}+{data_type}"] = Pattern(name, attrs)


class SqliteDB:
    """SQLite database class."""

    def __init__(self, path: str, fiona_aws_session=None):
        self.logger = get_logger(__name__)
        self.logger.debug(path)
        if not path.endswith(".sqlite"):
            raise ValueError(f"invalid extension for sqlite database: {path}")
        self.path = path
        self.fiona_aws_session = fiona_aws_session
        if self.fiona_aws_session:
            with fiona.Env(self.fiona_aws_session):
                self.layers = fiona.listlayers(self.path)
                self.reach_feats = gpd.read_file(self.path, layer="reach2d")
                self.subbasin_feats = gpd.read_file(self.path, layer="subbasin2d")
        else:
            self.layers = fiona.listlayers(self.path)
            self.reach_feats = gpd.read_file(self.path, layer="reach2d")
            self.subbasin_feats = gpd.read_file(self.path, layer="subbasin2d")

        # check consistent crs and assign crs to sqlite class
        if (
            self.reach_feats.crs != self.subbasin_feats.crs
        ):  # could also compare to coordinate system in the .basin file. once we parse that
            raise ValueError("coordinate system misalignment between subbasins and reaches")
        # else:
        #     self.crs = self.subbasin_feats.crs


class GridFile(BaseTextFile):
    """Class for parsing HEC-HMS grid files."""

    def __init__(self, path: str):
        if not path.endswith(".grid"):
            raise ValueError(f"invalid extension for Grid file: {path}")
        super().__init__(path)
        self.elements = ElementSet()
        self.scan_for_grids()

    def __repr__(self):
        """Representation of the HMSGridFile class."""
        return f"HMSGridFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Return grid manager name."""
        lines = self.content.splitlines()
        if not lines[0].startswith("Grid Manager: "):
            raise ValueError(f"unexpected first line: {lines[0]}")
        return lines[0][len("Grid Manager: ") :]

    def scan_for_grids(self):
        """Scan for all grids."""
        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith("Grid: "):
                name = line[len("Grid: ") :]
                grid_type = lines[i + 1][len("     Grid Type: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                self.elements[f"{name}+{grid_type}"] = Grid(f"{name}+{grid_type}", attrs)

    def remove_grid_type(self, grid_types: list[str]):
        """Remove given grid types."""
        new_elements = ElementSet()
        for name, g in self.elements.elements.items():
            if g.attrs["Grid Type"] not in grid_types:
                new_elements[name] = g
        self.elements = new_elements

    @property
    @lru_cache
    def grids(self):
        """Return grid elements."""
        return self.elements.get_element_type("Grid")


class GageFile(BaseTextFile):
    """Class for parsing HEC-HMS gage files."""

    def __init__(self, path: str):
        if not path.endswith(".gage"):
            raise ValueError(f"invalid extension for Gage file: {path}")
        super().__init__(path)
        self.elements = ElementSet()
        self.scan_for_gages()

    def __repr__(self):
        """Representation of the HMSGageFile class."""
        return f"HMSGageFile({self.path})"

    @property
    @lru_cache
    def name(self):
        """Return gage manager name."""
        lines = self.content.splitlines()
        if not lines[0].startswith("Gage Manager: "):
            raise ValueError(f"unexpected first line: {lines[0]}")
        return lines[0][len("Gage Manager: ") :]

    def scan_for_gages(self):
        """Search for all gages."""
        lines = self.content.splitlines()
        for i, line in enumerate(lines):
            if line.startswith("Gage: "):
                name = line[len("Gage: ") :]
                attrs = utils.parse_attrs(lines[i + 1 :])
                self.elements[name] = Gage(name, attrs)

    @property
    @lru_cache
    def gages(self):
        """Return gage elements."""
        return self.elements.get_element_type("Gage")



"""Utilities to extract specific timeseries from HEC-RAS."""

import os

import pandas as pd
from rashdf import RasPlanHdf
from rashdf.plan import RasPlanHdfError


def save_df_as_pq(df: pd.DataFrame, path: str):
    """Save DataFrame to parquet on local or on s3."""
    if "s3://" in path:
        df.to_parquet(path)
    else:
        dir_path = os.path.dirname(path)
        os.makedirs(dir_path, exist_ok=True)
        df.to_parquet(path)


def save_reference_lines(plan_hdf: RasPlanHdf, output_prefix: str) -> dict:
    """Process and save flow and water surface time series data for reference lines."""
    try:
        ref_line_ts = plan_hdf.reference_lines_timeseries_output()
    except RasPlanHdfError:
        return {}

    refln_paths = {}

    for refln_id in ref_line_ts.refln_id.values:
        refln_name = ref_line_ts.refln_name.sel(refln_id=refln_id).item()

        flow_df = ref_line_ts["Flow"].sel(refln_id=refln_id).to_series().reset_index()
        flow_df.columns = ["time", "flow"]

        wsel_df = ref_line_ts["Water Surface"].sel(refln_id=refln_id).to_series().reset_index()
        wsel_df.columns = ["time", "water_surface"]

        flow_path = f"{output_prefix}/ref_line={refln_name}/flow.pq"
        wsel_path = f"{output_prefix}/ref_line={refln_name}/wsel.pq"

        save_df_as_pq(flow_df, flow_path)
        save_df_as_pq(wsel_df, wsel_path)

        refln_paths[refln_name] = [flow_path, wsel_path]

    return refln_paths


def save_reference_points(plan_hdf: RasPlanHdf, output_prefix: str) -> dict:
    """Process and save velocity and water surface time series data for reference points."""
    try:
        ref_point_ts = plan_hdf.reference_points_timeseries_output()
    except RasPlanHdfError:
        return {}

    refpt_paths = {}

    for refpt_id in ref_point_ts.refpt_id.values:
        refpt_name = ref_point_ts.refpt_name.sel(refpt_id=refpt_id).item()

        velocity_df = ref_point_ts["Velocity"].sel(refpt_id=refpt_id).to_series().reset_index()
        velocity_df.columns = ["time", "velocity"]

        wsel_df = ref_point_ts["Water Surface"].sel(refpt_id=refpt_id).to_series().reset_index()
        wsel_df.columns = ["time", "water_surface"]

        velocity_path = f"{output_prefix}/ref_point={refpt_name}/velocity.pq"
        wsel_path = f"{output_prefix}/ref_point={refpt_name}/wsel.pq"

        save_df_as_pq(velocity_df, velocity_path)
        save_df_as_pq(wsel_df, wsel_path)

        refpt_paths[refpt_name] = [velocity_path, wsel_path]

    return refpt_paths


def save_bc_lines(plan_hdf: RasPlanHdf, output_prefix: str) -> dict:
    """Process and save velocity and water surface time series data for reference points."""
    try:
        bs_line_ts = plan_hdf.bc_lines_timeseries_output()
    except RasPlanHdfError:
        return {}

    bc_ln_paths = {}

    for bc_line_id in bs_line_ts.bc_line_id.values:
        bc_line_name = bs_line_ts.bc_line_name.sel(bc_line_id=bc_line_id).item()

        stage_df = bs_line_ts["Stage"].sel(bc_line_id=bc_line_id).to_series().reset_index()
        stage_df.columns = ["time", "stage"]

        flow_df = bs_line_ts["Flow"].sel(bc_line_id=bc_line_id).to_series().reset_index()
        flow_df.columns = ["time", "flow"]

        stage_path = f"{output_prefix}/bc_line={bc_line_name}/stage.pq"
        flow_path = f"{output_prefix}/bc_line={bc_line_name}/flow.pq"

        save_df_as_pq(stage_df, stage_path)
        save_df_as_pq(flow_df, flow_path)

        bc_ln_paths[bc_line_name] = [stage_path, flow_path]

    return bc_ln_paths



"""Create instances of assets."""

from pathlib import Path
from typing import Dict, Generic, Optional, Type, TypeVar

from pyproj import CRS
from pystac import Asset

from hecstac.common.logger import get_logger
from hecstac.hms.s3_utils import check_storage_extension
from hecstac.ras.utils import is_unc_path

T = TypeVar("T")  # Generic for asset file accessor classes


class GenericAsset(Asset, Generic[T]):
    """Provides a base structure for assets."""

    regex_parse_str: str = r""
    __roles__: list[str] = []
    __description__: str = ""
    __media_type__: Optional[str] = None
    __file_class__: T

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.description is None:
            self.description = self.__description__
        self._roles = []
        self._extra_fields = {}
        self.name = Path(self.href).name
        self.media_type = self.__media_type__
        self.logger = get_logger(__file__)

    @property
    def roles(self) -> list[str]:
        """Return roles with enforced values."""
        roles = self._roles
        for i in self.__roles__:
            if i not in roles:
                roles.append(i)
        return roles

    @roles.setter
    def roles(self, roles: list):
        self._roles = roles

    @property
    def extra_fields(self):
        """Return extra fields."""
        # boilerplate here, but overwritten in subclasses
        return self._extra_fields

    @extra_fields.setter
    def extra_fields(self, extra_fields: dict):
        """Set user-defined extra fields."""
        self._extra_fields = extra_fields

    @property
    def file(self) -> T:
        """Return cached file or instantiate new."""
        if not hasattr(self, "_file_obj"):
            if self.__file_class__:
                self._file_obj = self.__file_class__(self.href)
            else:
                raise AttributeError(f"No file class defined for {self.__class__.__name__}")
        return self._file_obj

    def name_from_suffix(self, suffix: str) -> str:
        """Generate a name by appending a suffix to the file stem."""
        return f"{self.stem}.{suffix}"

    @property
    def crs(self) -> CRS:
        """Get the authority code for the model CRS."""
        if self.ext.has("proj"):
            wkt2 = self.ext.proj.wkt2
            if wkt2 is None:
                return
            else:
                return CRS(wkt2)

    def __repr__(self):
        """Return string representation of the GenericAsset instance."""
        return f"<{self.__class__.__name__} name={self.name}>"

    def __str__(self):
        """Return string representation of assets name."""
        return f"{self.name}"


class AssetFactory:
    """Factory for creating HEC asset instances based on file extensions."""

    def __init__(self, extension_to_asset: Dict[str, Type[GenericAsset]]):
        """Initialize the AssetFactory with a mapping of file extensions to asset types and metadata."""
        self.extension_to_asset = extension_to_asset
        self._cache = {}  # Cache to store created assets

    def create_hms_asset(self, fpath: str, item_type: str = "model") -> Asset:
        """
        Create an asset instance based on the file extension.

        item_type: str

        The type of item to create. This is used to determine the asset class.
        Options are event or model.
        """
        if item_type not in ["event", "model"]:
            raise ValueError(f"Invalid item type: {item_type}, valid options are 'event' or 'model'.")

        # Check cache first
        cache_key = (fpath, item_type)
        if cache_key in self._cache:
            return self._cache[cache_key]

        file_extension = Path(fpath).suffix.lower()
        if file_extension == ".basin":
            asset_class = self.extension_to_asset.get(".basin").get(item_type)
        else:
            asset_class = self.extension_to_asset.get(file_extension, GenericAsset)

        asset = asset_class(href=fpath)
        asset.title = Path(fpath).name

        # Cache the created asset
        self._cache[cache_key] = asset

        return check_storage_extension(asset)

    def asset_from_dict(self, asset: Asset):
        """Create HEC asset given a base Asset and a map of file extensions dict."""
        fpath = asset.href

        # Check cache first
        if fpath in self._cache:
            return self._cache[fpath]

        for pattern, asset_class in self.extension_to_asset.items():
            if pattern.match(fpath):
                # logger.debug(f"Matched {pattern} for {Path(fpath).name}: {asset_class}")
                created_asset = asset_class.from_dict(asset.to_dict())
                # Overwrite the asset href with original filepath if it's a UNC path
                if is_unc_path(fpath):
                    created_asset.href = fpath

                # Cache the created asset
                self._cache[fpath] = created_asset

                return created_asset

        # Cache the generic asset if no match is found
        self._cache[fpath] = asset
        return asset



"""HMS FFRD Event Item Module."""

import json
from pystac import Asset, Item, Link
from hecstac.common.base_io import ModelFileReader
from shapely import to_geojson, union_all
from shapely.geometry import shape
import numpy as np
from datetime import datetime


class HMSEventItem(Item):
    """
    STAC Item subclass for representing an FFRD HMS event.

    This class builds a STAC Item from simulation metadata, with methods to create assets such as
    control files, basin files, and simulation outputs based on FEMA FFRD S3 file structure.
    """

    def __init__(
        self,
        realization: int,
        block_index: int,
        event_number: int,
        event_index_by_block: int,
        source_model_paths: list[str],
        proj_str: str,
    ) -> None:
        self.realization = realization
        self.block_index = block_index
        self.event_number = event_number
        self.event_index_by_block = event_index_by_block
        self.source_model_paths = source_model_paths
        self.source_model_items = []
        self.proj_str = proj_str
        for path in source_model_paths:
            ras_model_dict = json.loads(ModelFileReader(path).content)
            self.source_model_items.append(Item.from_dict(ras_model_dict))
        super().__init__(
            self._item_id,
            self._geometry,
            self._bbox,
            self._datetime,
            self._properties,
            href=None,
        )

    def build_ffrd_assets(
        self,
        basin_path: str = None,
        control_file_path: str = None,
        sst_dss_path: str = None,
        hdf_path: str = None,
        sim_data_path: str = None,
    ):
        """Build the assets for the STAC item.

        Parameters
        ----------
        basin_path : str, optional
            Path to the basin file, by default None
        control_file_path : str, optional
            Path to the control file, by default None
        sst_dss_path : str, optional
            Path to the complete sim SST DSS file, by default None
        hdf_path : str, optional
            Path to the excess precip HDF file, by default None
        sim_data_path : str, optional
            Path to the simulation data file, by default None
        """
        if sst_dss_path:
            self.add_sst_dss_asset(sst_dss_path)
        if hdf_path:
            self.add_hdf_file_asset(hdf_path)
        if basin_path:
            self.add_basin_file_asset(basin_path)
        if control_file_path:
            self.add_control_file_asset(control_file_path)
        if sim_data_path:
            self.add_sim_data_asset(sim_data_path)

    @property
    def _item_id(self) -> str:
        """Generate a unique item ID based on realization, block index, and event index."""
        if self.realization and self.block_index and self.event_index_by_block:
            return f"r{self.realization:02d}-b{self.block_index:04d}-e{self.event_index_by_block:02d}"
        return str(self.event_number)

    @property
    def _geometry(self) -> dict | None:
        """Calculate the geometry for the item based on source model items."""
        geometries = [shape(item.geometry) for item in self.source_model_items]
        return json.loads(to_geojson(union_all(geometries)))

    @property
    def _bbox(self) -> list[float]:
        """Calculate the bounding box for the item based on source model items."""
        if len(self.source_model_items) > 1:
            bboxes = np.array([item.bbox for item in self.source_model_items])
            return [
                float(bboxes[:, 0].min()),
                float(bboxes[:, 1].min()),
                float(bboxes[:, 2].max()),
                float(bboxes[:, 3].max()),
            ]
        return self.source_model_items[0].bbox

    @property
    def _datetime(self) -> datetime:
        """Use item creation time as the datetime."""
        return datetime.now()

    @property
    def _properties(self):
        """Build the properties for the STAC item."""
        return {
            "block_group": self.block_index,
            "event_id": self.event_number,
            "realization": self.realization,
            "proj:wkt2": self.proj_str,
            "data_time_source": "Item creation time",
        }

    def add_sst_dss_asset(self, full_dss_path: str):
        """Add complete sim SST DSS file as an asset."""
        self.add_asset(
            "complete_sim_output",
            Asset(href=full_dss_path, title="complete_sim_output", media_type="application/x-dss"),
        )

    def add_hdf_file_asset(self, full_hdf_path: str):
        """Add excess precip HDF file as an asset."""
        self.add_asset(
            "excess_precip",
            Asset(href=full_hdf_path, title="excess_precip", media_type="application/x-hdf"),
        )

    def add_basin_file_asset(self, full_basin_path: str):
        """Add basin file as an asset."""
        if not full_basin_path.endswith(".basin"):
            full_basin_path += ".basin"
        self.add_asset("basin", Asset(href=full_basin_path, title="basin", media_type="text/plain"))

    def add_control_file_asset(self, control_file_path: str):
        """Add control file as an asset."""
        self.add_asset("control", Asset(href=control_file_path, title="control", media_type="text/plain"))

    def add_sim_data_asset(self, sim_data_path: str):
        """Add simulation output data as an asset."""
        self.add_asset(
            "select_sim_output",
            Asset(
                href=sim_data_path,
                title="select_sim_output",
                description="Select time series extracted from hms simulation output.",
                media_type="application/x-parquet",
                extra_fields={"flow_time_series": "FLOW.pq", "base_flow_time_series": "FLOW-BASE.pq"},
            ),
        )

    def add_authoritative_model_link(self, item_href: str = None):
        """Add a link to the authoritative model(s). If item_href is provided, it links to that item; otherwise, it links to the source model paths."""
        if item_href:
            self.add_link(Link(rel="derived_from", target=item_href, title="Source Model"))
        else:
            for href in self.source_model_paths:
                self.add_link(Link(rel="derived_from", target=href, title="Source Model"))

    def add_storm_item_link(self, storm_item_href: str):
        """Add a link to the storm item."""
        self.add_link(Link(rel="derived_from", target=storm_item_href, title="Storm Item"))



"""Common utility functions."""

import json
import os
from pathlib import Path

from pystac import Item


def sanitize_catalog_assets(item: Item) -> Item:
    """Force the asset paths in the catalog to be relative to the item root."""
    item_dir = Path(item.pm.item_dir).resolve()

    for _, asset in item.assets.items():
        asset_path = Path(asset.href).resolve()

        if asset_path.is_relative_to(item_dir):
            asset.href = str(asset_path.relative_to(item_dir))
        else:
            asset.href = (
                str(asset_path.relative_to(item_dir.parent))
                if item_dir.parent in asset_path.parents
                else str(asset_path)
            )

    return item


def load_config(config_input: str) -> list[dict[str, str]]:
    """
    Load a HECSTAC workflow configuration from a JSON string or a file path.

    This function is used to load configurations for HECSTAC workflows, such as creating RAS STAC items.
    If `config_input` is a path to a local file, the file is read and parsed as JSON.
    If `config_input` is a JSON string, it is parsed directly.
    The configuration must be either a dictionary or a list of dictionaries.

    Args:
        config_input (str): Path to a JSON config file or a JSON string.

    Returns
    -------
        list[dict]: List of configuration dictionaries.

    Example:
        Single config (as string or file):
            {
                "ras_project_path": "...",
                "output_prefix": "..."
            }
        Multiple configs:
            [
                {"ras_project_path": "...", "output_prefix": "..."},
                {"ras_project_path": "...", "output_prefix": "..."}
            ]
    """
    # add option for reading local file
    if os.path.isfile(config_input):
        with open(config_input, "r") as f:
            config_data = json.load(f)
    else:
        config_data = json.loads(config_input)

    # always pass config as a list
    if isinstance(config_data, dict):
        return [config_data]
    elif isinstance(config_data, list):
        return config_data
    else:
        raise ValueError(f"Config must be a JSON object or list of objects not {type(config_data)}: {config_data}")



"""Constants."""

import datetime
import json

from shapely import Polygon, box, to_geojson

SCHEMA_URI = (
    "https://raw.githubusercontent.com/fema-ffrd/hecstac/refs/heads/port-ras-stac/hecstac/ras/extension/schema.json"
)

NULL_DATETIME = datetime.datetime(9999, 9, 9)
NULL_GEOMETRY = Polygon([(0, 0), (0, 1), (1, 1), (1, 0)])
NULL_STAC_GEOMETRY = json.loads(to_geojson(NULL_GEOMETRY))
NULL_BBOX = box(0, 0, 1, 1)
NULL_STAC_BBOX = NULL_BBOX.bounds
PLACEHOLDER_ID = "id"



"""Class for event items."""

import json
import logging
import re
import os
from datetime import datetime
from pathlib import Path
from typing import List

from functools import cached_property
import numpy as np
import pandas as pd
from pystac import Asset, Item, Link
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.storage import StorageExtension
from rashdf import RasPlanHdf
from shapely import simplify, to_geojson, union_all
from shapely.geometry import shape

from hecstac.common.base_io import ModelFileReader
from hecstac.common.asset_factory import AssetFactory
from hecstac.common.logger import get_logger
from hecstac.hms.assets import HMS_EXTENSION_MAPPING
from hecstac.ras.assets import RAS_EXTENSION_MAPPING
from hecstac.events.ts_utils import save_bc_lines, save_reference_lines, save_reference_points

logger = get_logger(__name__)

PARQUET_MEDIA_TYPE = "application/x-parquet"


class FFRDEventItem(Item):
    """Class for event items."""

    FFRD_REALIZATION = "FFRD:realization"
    FFRD_BLOCK_GROUP = "FFRD:block_group"
    FFRD_EVENT = "FFRD:event"

    def __init__(
        self,
        ras_simulation_files: list,
        source_model_paths: list = None,
        event_id: str = None,
        realization: str = None,
        block_group: str = None,
        hms_simulation_files: list = None,
    ) -> None:
        self.realization = realization
        self.block_group = block_group
        self.source_model_paths = source_model_paths
        self.source_model_items = []
        self.stac_extensions = None
        self.hms_simulation_files = hms_simulation_files
        self.ras_simulation_files = ras_simulation_files
        self.event_id = event_id or self._event_id_from_plan_hdf()
        self.hms_factory = AssetFactory(HMS_EXTENSION_MAPPING)
        self.ras_factory = AssetFactory(RAS_EXTENSION_MAPPING)

        if source_model_paths is not None:
            for path in source_model_paths:
                ras_model_dict = json.loads((ModelFileReader(path).content))
                self.source_model_items.append(Item.from_dict(ras_model_dict))

        super().__init__(
            self._item_id,
            self._geometry,
            self._bbox,
            self._datetime,
            self._properties,
            href=None,
        )

        for fpath in self.hms_simulation_files or []:
            self.add_hms_asset(fpath, item_type="event")

        for fpath in self.ras_simulation_files:
            self.add_ras_asset(fpath)

        self._register_extensions()
        self._add_model_links()

    def _register_extensions(self) -> None:
        ProjectionExtension.add_to(self)
        StorageExtension.add_to(self)

    def _add_model_links(self) -> None:
        """Add links to the model items."""
        for item in self.source_model_items:
            logger.info(f"Adding link from source model item: {item.id}")
            link = Link(
                rel="derived_from",
                target=item,
                title=f"Source Models: {item.id}",
            )
            self.add_link(link)

    def _event_id_from_plan_hdf(self):
        return self.plan_hdf.get_plan_info_attrs()["Plan Name"]

    @cached_property
    def plan_hdf(self):
        """Returns the main RAS plan HDF file from the ras simulation files."""
        plan_hdf_pattern = re.compile(r"p\d{2}\.hdf$")
        for ras_file in self.ras_simulation_files:
            if plan_hdf_pattern.search(ras_file):
                logger.info(f"Using {ras_file} as main plan hdf file.")
                return RasPlanHdf.open_uri(
                    ras_file, fsspec_kwargs={"default_cache_type": "blockcache", "default_block_size": 10**5}
                )

        raise ValueError("No plan HDF file found.")

    @property
    def _item_id(self) -> str:
        """The event id for the FFRD Event STAC item."""
        if self.realization and self.block_group and self.event_id:
            return f"{self.realization}-{self.block_group}-{self.event_id}"
        else:
            return self.event_id

    @property
    def _properties(self):
        """Properties for the HMS STAC item."""
        properties = {}
        properties[self.FFRD_EVENT] = self.event_id

        if self.realization:
            properties[self.FFRD_REALIZATION] = self.realization
        if self.block_group:
            properties[self.FFRD_BLOCK_GROUP] = self.block_group
        return properties

    @property
    def _geometry(self) -> dict | None:
        """Geometry of the FFRD Event STAC item. Union of all basins in the FFRD Event items."""
        if len(self.source_model_items) == 0:
            areas = self.plan_hdf.mesh_areas()

            if areas.crs is not None:
                areas = areas.to_crs("EPSG:4326")
            else:
                logger.warning("No CRS found with mesh areas, using geometry without reprojection.")

            unioned_geometry = union_all(areas.geometry)
            unioned_geometry = simplify(unioned_geometry, 0.001)
            return json.loads(to_geojson(unioned_geometry))
        else:
            geometries = [shape(item.geometry) for item in self.source_model_items]
            return json.loads(to_geojson(union_all(geometries)))

    @property
    def _datetime(self) -> datetime:
        """The datetime for the FFRD Event STAC item."""
        return datetime.now()

    @property
    def _bbox(self) -> list[float]:
        """Bounding box of the FFRD Event STAC item."""
        if len(self.source_model_items) == 0:
            areas = self.plan_hdf.mesh_areas()
            if areas.crs is not None:
                areas = areas.to_crs("EPSG:4326")
            bounds = areas.total_bounds
            return [float(b) for b in bounds]
        elif len(self.source_model_items) > 1:
            bboxes = np.array([item.bbox for item in self.source_model_items])
            bboxes = [bboxes[:, 0].min(), bboxes[:, 1].min(), bboxes[:, 2].max(), bboxes[:, 3].max()]
            return [float(i) for i in bboxes]
        else:
            return self.source_model_items[0].bbox

    def add_hms_asset(self, fpath: str, item_type: str = "event") -> None:
        """Add an asset to the FFRD Event STAC item."""
        if os.path.exists(fpath):
            logger.info(f"Adding asset: {fpath}")
            asset = self.hms_factory.create_hms_asset(fpath, item_type=item_type)
            if asset is not None:
                self.add_asset(asset.title, asset)

    def add_ras_asset(self, fpath: str) -> None:
        """Add an asset to the FFRD Event STAC item."""
        logger.info(f"Adding asset: {fpath}")
        asset = Asset(href=fpath, title=Path(fpath).name)
        asset = self.ras_factory.asset_from_dict(asset)
        if asset is not None:
            self.add_asset(asset.title, asset)

    def _add_ts_assets_from_dict(self, asset_dict: dict, description: str):
        """Add time data as item assets."""
        for name, paths in asset_dict.items():
            for path in paths:
                var_type = path.split("/")[-1].split(".")[0]
                title = f"{name}-{var_type}"
                self.add_asset(
                    title,
                    Asset(
                        href=path,
                        title=title,
                        description=description,
                        media_type=PARQUET_MEDIA_TYPE,
                        roles=["data"],
                    ),
                )

    def add_ts_assets(self, prefix: str):
        """
        Extract and add time series assets from the plan HDF file to the STAC item. Seperates data by element and data type (stage and flow) and saves as Parquet files for each individual element and data type.

        This includes:
        - Boundary condition line time series
        - Reference line time series
        - Reference point time series

        Args:
            prefix (str): S3 or local path prefix where the time series assets will be stored.
        """
        plan_hdf = self.plan_hdf

        # Boundary condition lines
        logger.info("Processing BC Line time series data.")
        bc_ln_paths = save_bc_lines(plan_hdf, prefix)
        if bc_ln_paths:
            self._add_ts_assets_from_dict(bc_ln_paths, "Parquet containing bc line time series data.")
        else:
            logger.info("No bc lines found.")

        # Reference lines
        logger.info("Processing reference line time series data.")
        refln_paths = save_reference_lines(plan_hdf, prefix)
        if refln_paths:
            self._add_ts_assets_from_dict(refln_paths, "Parquet containing reference line time series data.")
        else:
            logger.info("No reference lines found.")

        # Reference points
        logger.info("Processing reference point time series data.")
        refpt_paths = save_reference_points(plan_hdf, prefix)
        if refpt_paths:
            self._add_ts_assets_from_dict(refpt_paths, "Parquet containing reference point time series data.")
        else:
            logger.info("No reference points found.")

    def add_flow_ts_asset(self, output_path: str):
        """Extract flow time series from boundary condition lines and reference lines. All data is combined into a single Parquet file and saved as an asset."""
        plan_hdf = self.plan_hdf

        bcline = plan_hdf.bc_lines_flow(use_names=True)
        ref_ln_flow = plan_hdf.reference_lines_flow(use_names=True)

        all_flows = pd.concat([bcline, ref_ln_flow], axis=1)

        # BC and reference lines can sometimes have same point, remove duplicates
        all_flows = all_flows.loc[:, ~all_flows.columns.duplicated(keep="first")]

        all_flows.to_parquet(output_path)

        self.add_asset(
            "flow_data",
            Asset(
                href=output_path,
                title="flow_data",
                description="Parquet containing time series flow data for reference lines and bc lines.",
                media_type=PARQUET_MEDIA_TYPE,
                roles=["data"],
            ),
        )

    def add_stage_ts_asset(self, output_path: str):
        """Extract stage time series from boundary condition lines, reference lines, and reference points. All data is combined into a single Parquet file and saved as an asset."""
        plan_hdf = self.plan_hdf

        bc_line_ts = plan_hdf.bc_lines_timeseries_output()
        ref_line_ts = plan_hdf.reference_lines_timeseries_output()

        bc_line_stage = (
            bc_line_ts["Stage"].to_dataframe().reset_index().pivot(index="time", columns="bc_line_name", values="Stage")
        )
        ref_line_stage = (
            ref_line_ts["Water Surface"]
            .to_dataframe()
            .reset_index()
            .pivot(index="time", columns="refln_name", values="Water Surface")
        )

        ref_pnt_stage = plan_hdf.reference_points_stage(use_names=True)

        all_stages = pd.concat([bc_line_stage, ref_pnt_stage, ref_line_stage], axis=1)

        # BC and reference lines can sometimes have same point, remove duplicates
        all_stages = all_stages.loc[:, ~all_stages.columns.duplicated(keep="first")]

        all_stages.to_parquet(output_path)

        self.add_asset(
            "stage_data",
            Asset(
                href=output_path,
                title="stage_data",
                description="Parquet containing time series stage data for reference lines, reference points, and bc lines.",
                media_type=PARQUET_MEDIA_TYPE,
                roles=["data"],
            ),
        )



"""Features developed duirng R&D for consideration in the hecstac package."""

import re
from pathlib import Path

import boto3

from hecstac.common.logger import get_logger
from hecstac.common.s3_utils import (
    list_keys_regex,
    metadata_to_s3,
    qc_results_to_excel_s3,
    verify_file_exists,
)
from hecstac.ras.item import RASModelItem
from rasqc.check import check
from rasqc.utils import summarize_results


def build_s3_path(
    bucket: str,
    prefix: str,
    model_name: str = "",
    metadata_part: str = "metadata",
    suffix: str = "",
    extension: str = "",
) -> str:
    """Make consistent s3 paths for metadata."""
    path = f"s3://{bucket}/{prefix}/{metadata_part}"
    if model_name:
        path += f"/{model_name}"
    if suffix:
        path += f"_{suffix}"
    if extension:
        path += f".{extension.lstrip('.')}"
    return path


class RASModelCalibrationError(Exception):
    """Custom exception for calibration processing errors."""

    pass


class RASModelCalibrationChecker:
    """Utility for streamlining QC checks for FFRD."""

    def __init__(
        self,
        s3_client: boto3.Session.client,
        bucket: str,
        prefix: str,
        ras_model_name: str = None,
        skip_hdf_files: bool = False,
        crs=None,
    ):
        self.s3_client = s3_client
        self.bucket = bucket
        self.prefix = prefix
        self.crs = crs
        self.skip_hdf_files = skip_hdf_files

        s3_parts = Path(prefix).parts
        self.ras_model_name = ras_model_name or s3_parts[-1]
        self.ras_project_key = f"{self.prefix}/{self.ras_model_name}.prj"
        self.item_id = Path(self.ras_project_key).stem
        self.logger = get_logger(__name__)

    def parse_files(self):
        """Parse s3 keys to identify RAS files."""
        verify_file_exists(bucket=self.bucket, key=self.ras_project_key, s3_client=self.s3_client)

        ras_prefix = f"{self.prefix}/{self.ras_model_name}"
        ras_files = list_keys_regex(s3_client=self.s3_client, bucket=self.bucket, prefix_includes=ras_prefix)
        ras_files = [f"s3://{self.bucket}/{f}" for f in ras_files]

        if self.skip_hdf_files:
            ras_files = [f for f in ras_files if not f.endswith(".hdf")]

        return ras_files

    def create_item(self, ras_files):
        """Create a STAC Item from a RAS model."""
        ras_item = RASModelItem.from_prj(
            build_s3_path(self.bucket, self.prefix, self.ras_model_name),
            crs=self.crs,
            assets=ras_files,
        )

        thumbnail_dst = build_s3_path(
            self.bucket,
            self.prefix,
        )

        ras_item.add_model_thumbnails(
            ["mesh_areas", "breaklines", "bc_lines"],
            thumbnail_dest=thumbnail_dst,
        )

        item_json = build_s3_path(
            self.bucket,
            self.prefix,
            self.ras_model_name,
            extension="json",
        )
        ras_item.set_self_href(item_json)
        ras_item.validate()

        return ras_item

    def upload_metadata(self, ras_item):
        """Upload metadata output."""
        self.logger.debug("upload_metadata starting")
        metadata_to_s3(
            bucket=self.bucket,
            prefix=self.prefix,
            model_name=self.ras_model_name,
            s3_client=self.s3_client,
            item=ras_item,
        )
        self.logger.info("upload_metadata complete")

    def run_qc(self, ras_item):
        """Perform QC check and upload to s3."""
        qc_results = summarize_results(check(ras_item, check_suite="ras_stac_ffrd"))
        self.logger.debug("run_qc results computed")
        qc_results_path = build_s3_path(
            self.bucket,
            self.prefix,
            self.ras_model_name,
            suffix="qc-results",
            extension="xlsx",
        )
        qc_results_to_excel_s3(qc_results, qc_results_path)
        self.logger.info("run_qc completed upload")



"""HEC-RAS STAC Item class."""

import json
import logging
import os
from datetime import datetime
from functools import lru_cache
from pathlib import Path

import contextily as ctx
import matplotlib.pyplot as plt
import numpy as np
import requests
from pystac import Asset, Item
from pystac.extensions.projection import ProjectionExtension
from pystac.extensions.storage import StorageExtension
from shapely import to_geojson, unary_union

from hecstac.common.asset_factory import AssetFactory
from hecstac.common.logger import get_logger
from hecstac.common.path_manager import LocalPathManager
from hecstac.hms.assets import HMS_EXTENSION_MAPPING
from hecstac.hms.parser import BasinFile, ProjectFile
from hecstac.ras.consts import NULL_DATETIME, NULL_STAC_BBOX, NULL_STAC_GEOMETRY

logger = get_logger(__name__)


class HMSModelItem(Item):
    """An object representation of a HEC-HMS model."""

    PROJECT = "hms:project"
    PROJECT_TITLE = "hms:project_title"
    MODEL_UNITS = "hms:unit system"
    MODEL_GAGES = "hms:gages"
    PROJECT_VERSION = "hms:version"
    PROJECT_DESCRIPTION = "hms:description"
    PROJECT_UNITS = "hms:unit_system"
    SUMMARY = "hms:summary"

    def __init__(self, *args, **kwargs):
        """Add a few default properties to the base class."""
        super().__init__(*args, **kwargs)
        self.simplify_geometry = True

    @classmethod
    def from_prj(
        cls, hms_project_file, item_id: str, simplify_geometry: bool = True, assets: list = None, asset_dir: str = None
    ):
        """
        Create an `HMSModelItem` from a HEC-HMS project file.

        Parameters
        ----------
        hms_project_file : str
            Path to the HEC-HMS project file (.hms).
        item_id : str
            Unique item ID for the STAC item.
        simplify_geometry : bool, optional
            Whether to simplify geometry. Defaults to True.

        Returns
        -------
        stac : HMSModelItem
            An instance of the class representing the STAC item.
        """
        pm = LocalPathManager(Path(hms_project_file).parent)
        pf = ProjectFile(hms_project_file, assert_uniform_version=False)
        # Create GeoJSON and Thumbnails
        # cls._check_files_exists(cls, pf.files + pf.rasters)

        # To access instance methods
        temp_instance = cls(
            Path(hms_project_file).stem,
            NULL_STAC_GEOMETRY,
            NULL_STAC_BBOX,
            NULL_DATETIME,
            {"hms_project_file": hms_project_file},
            href="",
            assets={},
        )
        temp_instance.pm = pm
        temp_instance.simplify_geometry = simplify_geometry

        geojson_paths = temp_instance.write_element_geojsons(pf.basins, pm, pf, asset_dir)
        thumbnail_paths = temp_instance.make_thumbnails(pf.basins, pm, asset_dir)

        # Collect all assets
        if not assets:
            href = pm.item_path(item_id)
            assets = {Path(i).name: Asset(i) for i in pf.files + pf.rasters + geojson_paths + thumbnail_paths}
        else:
            href = hms_project_file
            assets = {Path(i).name: Asset(i, Path(i).name) for i in assets}
        # Create the STAC Item
        stac = cls(
            Path(hms_project_file).stem,
            NULL_STAC_GEOMETRY,
            NULL_STAC_BBOX,
            NULL_DATETIME,
            {"hms_project_file": hms_project_file},
            href=href,
            assets=assets,
        )
        stac.pm = pm
        stac.simplify_geometry = simplify_geometry

        stac._register_extensions()
        return stac

    def _register_extensions(self) -> None:
        ProjectionExtension.add_to(self)
        StorageExtension.add_to(self)

    @property
    def hms_project_file(self) -> str:
        """Get the path to the HEC-HMS .hms file."""
        return self._properties.get("hms_project_file")

    @property
    @lru_cache
    def factory(self) -> AssetFactory:
        """Return AssetFactory for this item."""
        return AssetFactory(HMS_EXTENSION_MAPPING)

    @property
    @lru_cache
    def pf(self) -> ProjectFile:
        """Get a ProjectFile instance for the HMS Model .hms file."""
        return ProjectFile(self.hms_project_file)

    @property
    def properties(self) -> dict:
        """Properties for the HMS STAC item."""
        properties = self._properties
        properties[self.PROJECT] = f"{self.pf.name}.hms"
        properties[self.PROJECT_TITLE] = self.pf.name
        properties[self.PROJECT_VERSION] = self.pf.attrs["Version"]
        properties[self.PROJECT_DESCRIPTION] = self.pf.attrs.get("Description")

        # Get data from the first basin
        properties[self.MODEL_UNITS] = self.pf.basins[0].attrs["Unit System"]
        properties[self.MODEL_GAGES] = self.pf.basins[0].gages
        properties["proj:code"] = self.pf.basins[0].epsg

        if self.pf.basins[0].epsg is None:
            logger.warning("No EPSG code found in basin file.")

        properties["proj:wkt"] = self.pf.basins[0].wkt
        properties[self.SUMMARY] = self.pf.file_counts

        return properties

    @properties.setter
    def properties(self, properties: dict):
        """Set properties."""
        self._properties = properties

    @property
    def geometry_assets(self) -> list[BasinFile]:
        """Return list of basin geometry assets."""
        return self.pf.basins

    @property
    def geometry(self) -> dict:
        """Return footprint of the model as a GeoJSON."""
        if not self.geometry_assets:
            return NULL_STAC_GEOMETRY

        geometries = [
            b.basin_geom.simplify(0.001) if self.simplify_geometry else b.basin_geom for b in self.geometry_assets
        ]
        unioned_geometry = unary_union(geometries)

        return json.loads(to_geojson(unioned_geometry))

    @property
    def bbox(self) -> list[float]:
        """Bounding box of the HMS model."""
        if not self.geometry_assets:
            return NULL_STAC_BBOX

        bboxes = np.array([b.bbox(4326) for b in self.geometry_assets])
        return [float(i) for i in [bboxes[:, 0].min(), bboxes[:, 1].min(), bboxes[:, 2].max(), bboxes[:, 3].max()]]

    @property
    def datetime(self) -> datetime:
        """The datetime for the HMS STAC item."""
        date = datetime.strptime(self.pf.basins[0].header.attrs["Last Modified Date"], "%d %B %Y")
        time = datetime.strptime(self.pf.basins[0].header.attrs["Last Modified Time"], "%H:%M:%S").time()
        return datetime.combine(date, time)

    def _check_files_exists(self, files: list[str]):
        """Ensure the files exists. If they don't raise an error."""
        for file in files:
            if not os.path.exists(file):
                logger.warning(f"File not found {file}")

    def make_thumbnails(
        self, basins: list[BasinFile], pm: LocalPathManager, asset_dir: str, overwrite: bool = False
    ) -> list[str]:
        """Create a PNG thumbnail for each basin."""
        thumbnail_paths = []

        for bf in basins:
            png_name = f"{bf.name}.png".replace(" ", "_").replace("-", "_")
            if asset_dir:
                os.makedirs(asset_dir, exist_ok=True)
                thumbnail_path = os.path.join(asset_dir, png_name)
            else:
                thumbnail_path = pm.derived_item_asset(png_name)

            if not overwrite and os.path.exists(thumbnail_path):
                logger.info(f"Thumbnail for basin `{bf.name}` already exists. Skipping creation.")
            else:
                logger.info(f"{'Overwriting' if overwrite else 'Creating'} thumbnail for basin `{bf.name}`")
                fig = self.make_thumbnail(gdfs=bf.hms_schematic_2_gdfs)
                fig.savefig(thumbnail_path)
                fig.clf()
            thumbnail_paths.append(thumbnail_path)

        return thumbnail_paths

    def write_element_geojsons(
        self, basins: list[BasinFile], pm: LocalPathManager, pf, asset_dir: str, overwrite: bool = False
    ):
        """Write the HMS elements (Subbasins, Juctions, Reaches, etc.) to geojson."""
        geojson_paths = []
        for element_type in basins[0].elements.element_types:
            if asset_dir:
                os.makedirs(asset_dir, exist_ok=True)
                path = os.path.join(asset_dir, f"{element_type}.geojson")
            else:
                path = pm.derived_item_asset(f"{element_type}.geojson")
            if not overwrite and os.path.exists(path):
                logger.info(f"Geojson for {element_type} already exists. Skipping creation.")
            else:
                logger.info(f"Creating geojson for {element_type}")
                gdf = pf.basins[0].feature_2_gdf(element_type).to_crs(4326)
                # logger.debug(gdf.columns)
                keep_columns = ["name", "geometry", "Last Modified Date", "Last Modified Time", "Number Subreaches"]
                gdf = gdf[[col for col in keep_columns if col in gdf.columns]]
                gdf.to_file(path)
            geojson_paths.append(path)

        return geojson_paths

    def add_asset(self, key, asset):
        """Subclass asset then add."""
        subclass = self.factory.asset_from_dict(asset)
        if subclass is None:
            return
        return super().add_asset(key, subclass)

    def make_thumbnail(self, gdfs: dict):
        """Create a png from the geodataframes (values of the dictionary). The dictionary keys are used to label the layers in the legend."""
        cdict = {
            "Subbasin": "black",
            "Reach": "blue",
            "Junction": "red",
            "Source": "black",
            "Sink": "green",
            "Reservoir": "cyan",
            "Diversion": "black",
        }
        crs = gdfs["Subbasin"].crs
        fig, ax = plt.subplots(1, 1, figsize=(6, 6))
        # Add data
        for layer in gdfs.keys():
            if layer in cdict.keys():
                if layer == "Subbasin":
                    gdfs[layer].plot(ax=ax, edgecolor=cdict[layer], linewidth=1, label=layer, facecolor="none")
                elif layer == "Junction":
                    gdfs[layer].plot(ax=ax, color=cdict[layer], label=layer, markersize=25)
                else:
                    gdfs[layer].plot(ax=ax, color=cdict[layer], linewidth=1, label=layer, markersize=5)
        try:
            ctx.add_basemap(ax, crs=crs, source=ctx.providers.USGS.USTopo)
        except requests.exceptions.HTTPError:
            try:
                ctx.add_basemap(ax, crs=crs, source=ctx.providers.Esri.WorldStreetMap)
            except requests.exceptions.HTTPError:
                ctx.add_basemap(ax, crs=crs, source=ctx.providers.OpenStreetMap.Mapnik)

        # Format
        # ax.legend()
        ax.set_xticks([])
        ax.set_yticks([])
        fig.tight_layout()
        return fig

    @geometry.setter
    # Prevent external modification of dynamically generated geometry property
    def geometry(self, value):
        pass

    @bbox.setter
    # Prevent external modification of dynamically generated bbox property
    def bbox(self, value):
        pass

    @datetime.setter
    # Prevent external modification of dynamically generated datetime property
    def datetime(self, value):
        pass



"""Utility functions for the hecstac ras module."""

import logging
import os
import re
from functools import wraps
from io import BytesIO
from pathlib import Path
from typing import Callable

import contextily as ctx
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
from pyproj import CRS
from shapely import lib
from shapely.errors import UnsupportedGEOSVersionError
from shapely.geometry import LineString, MultiPoint, Point

from hecstac.common.base_io import ModelFileReader
from hecstac.common.logger import get_logger
from hecstac.common.s3_utils import save_bytes_s3

logger = get_logger(__name__)


def export_thumbnail(layers: list[Callable], title: str, crs: CRS, filepath: str):
    """Generate a thumbnail and save it."""
    fig, ax = plt.subplots(figsize=(12, 12))

    # Add data
    legend_handles = []
    for layer in layers:
        try:
            legend_handles += layer(ax)
        except Exception as e:
            logger.warning(f"Warning: Failed to process layer '{layer}' for thumbnail at {filepath}: {e}")

    # Add OpenStreetMap basemap
    try:
        ctx.add_basemap(ax, crs=crs, source=ctx.providers.OpenStreetMap.Mapnik)
    except Exception as e:
        logger.warning(f"Warning: Failed to add basemap for {filepath}: {e}")

    # Formatting
    ax.set_title(title)
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")
    ax.legend(handles=legend_handles, loc="center left", bbox_to_anchor=(1, 0.5))
    fig.tight_layout()

    # Save
    if filepath.startswith("s3://"):
        img_data = BytesIO()
        fig.savefig(img_data, format="png", bbox_inches="tight")
        img_data.seek(0)
        save_bytes_s3(img_data, filepath)
    else:
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        fig.savefig(filepath, dpi=80, bbox_inches="tight")

    # Close fig
    plt.close(fig)


def find_model_files(ras_prj: str) -> list[str]:
    """Find all files with the same base name and return absolute paths."""
    ras_prj = Path(ras_prj).resolve()
    parent = ras_prj.parent
    stem = ras_prj.stem
    return [str(i.resolve()) for i in parent.glob(f"{stem}*")]


def is_unc_path(file_path: str):
    """Check if a path is a Universal Naming Convention (UNC) network path."""
    if file_path.startswith("//") or file_path.startswith("\\\\"):
        return True
    else:
        return False


def is_ras_prj(url: str) -> bool:
    """Check if a file is a HEC-RAS project file."""
    file_str = ModelFileReader(url).content
    if "Proj Title" in file_str.split("\n")[0]:
        return True
    else:
        return False


def search_contents(
    lines: list[str],
    search_string: str,
    token: str = "=",
    expect_one: bool = True,
    require_one: bool = True,
    regex: bool = False,
) -> list[str] | str:
    """Split a line by a token and returns the second half of the line if the search_string is found in the first half.

    The regex option assumes that the token is included in the regex.
    """
    if regex:
        matches = lambda x: re.match(search_string, x)
    else:
        matches = lambda x: f"{search_string}{token}" in x
    results = []
    for line in lines:
        if matches(line):
            val = line.split(token)[1]
            if val != "":
                results.append(val)

    if expect_one and len(results) > 1:
        raise ValueError(f"expected 1 result for {search_string}, got {len(results)} results")
    elif require_one and len(results) == 0:
        raise ValueError(f"1 result for {search_string} is required, no results found")
    elif expect_one and len(results) == 1:
        return results[0]
    else:
        return results


def handle_spaces(line: str, lines: list[str]):
    """Handle spaces in the line."""
    if line in lines:
        return line
    elif handle_spaces_arround_equals(line.rstrip(" "), lines):
        return handle_spaces_arround_equals(line.rstrip(" "), lines)
    elif handle_spaces_arround_equals(line + " ", lines) in lines:
        return handle_spaces_arround_equals(line + " ", lines)
    else:
        raise ValueError(f"line: {line} not found in lines")


def handle_spaces_arround_equals(line: str, lines: list[str]) -> str:
    """Handle spaces in the line."""
    if line in lines:
        return line
    elif "= " in line:
        if line.replace("= ", "=") in lines:
            return line.replace("= ", "=")
    else:
        return line.replace("=", "= ")


def text_block_from_start_end_str(
    start_str: str, end_strs: list[str], lines: list, additional_lines: int = 0
) -> list[str]:
    """Search for an exact match to the start_str and return all lines from there to a line that contains the end_str."""
    start_str = handle_spaces(start_str, lines)

    start_index = lines.index(start_str)
    end_index = len(lines)
    for line in lines[start_index + 1 :]:
        if end_index != len(lines):
            break
        for end_str in end_strs:
            if end_str in line:
                end_index = lines.index(line) + additional_lines
                break
    return lines[start_index:end_index]


def text_block_from_start_str_to_empty_line(start_str: str, lines: list) -> list[str]:
    """Search for an exact match to the start_str and return all lines from there to the next empty line."""
    start_str = handle_spaces(start_str, lines)
    results = []
    in_block = False
    for line in lines:
        if line == start_str:
            in_block = True
            results.append(line)
            continue

        if in_block:
            if line == "":
                results.append(line)
                return results
            else:
                results.append(line)
    return results


def text_block_from_start_str_length(start_str: str, number_of_lines: int, lines: list[str]) -> list[str]:
    """Search for an exact match to the start token and return a number of lines equal to number_of_lines."""
    start_str = handle_spaces(start_str, lines)
    results = []
    in_block = False
    for line in lines:
        if line == start_str:
            in_block = True
            continue
        if in_block:
            if len(results) >= number_of_lines:
                return results
            else:
                results.append(line)


def data_pairs_from_text_block(lines: list[str], width: int) -> list[tuple[float, float]]:
    """Split lines at given width to get paired data string. Split the string in half and convert to tuple of floats."""
    pairs = []
    for line in lines:
        if line == "               .               .":
            continue
        for i in range(0, len(line), width):
            x = line[i : int(i + width / 2)]
            y = line[int(i + width / 2) : int(i + width)]
            pairs.append((float(x), float(y)))

    return pairs


def delimited_pairs_to_lists(lines: list[str]) -> tuple[list[float], list[float]]:
    """Extract subdivisions from the manning's text block."""
    stations = []
    mannings = []
    for line in lines:
        pairs = line.split("       0")
        for p in pairs[:-1]:
            station = float(p[:8])
            n = float(p[8:])
            stations.append(station)
            mannings.append(n)
    return (stations, mannings)


def data_triplets_from_text_block(lines: list[str], width: int) -> list[tuple[float]]:
    """Split lines at given width to get paired data string. Split the string in half and convert to tuple of floats."""
    pairs = []
    for line in lines:
        for i in range(0, len(line), width):
            x = line[i : int(i + width / 3)]
            y = line[int(i + width / 3) : int(i + (width * 2 / 3))]
            z = line[int(i + (width * 2 / 3)) : int(i + (width))]
            pairs.append((float(x), float(y), float(z)))

    return pairs


def check_xs_direction(cross_sections: gpd.GeoDataFrame, reach: LineString):
    """Return only cross sections that are drawn right to left looking downstream."""
    logger = get_logger(__name__)
    river_reach_rs = []
    for _, xs in cross_sections.iterrows():
        try:
            point = reach.intersection(xs["geometry"])
            point = validate_point(point)
            xs_rs = reach.project(point)

            offset = xs.geometry.offset_curve(-1)
            if reach.intersects(offset):  # if the offset line intersects then use this logic
                point = reach.intersection(offset)
                point = validate_point(point)

                offset_rs = reach.project(point)
                if xs_rs > offset_rs:
                    river_reach_rs.append(xs["river_reach_rs"])
            else:  # if the original offset line did not intersect then try offsetting the other direction and applying
                # the opposite stationing logic; the orginal line may have gone beyound the other line.
                offset = xs.geometry.offset_curve(1)
                point = reach.intersection(offset)
                point = validate_point(point)

                offset_rs = reach.project(point)
                if xs_rs < offset_rs:
                    river_reach_rs.append(xs["river_reach_rs"])

        except IndexError as e:
            logger.debug(
                f"cross section does not intersect river-reach: {xs['river']} {xs['reach']} {xs['river_station']}: error: {e}"
            )
            continue
    return cross_sections.loc[cross_sections["river_reach_rs"].isin(river_reach_rs)]


def validate_point(geom):
    """Validate that point is of type Point. If Multipoint or Linestring create point from first coordinate pair."""
    if isinstance(geom, Point):
        return geom
    elif isinstance(geom, MultiPoint):
        return geom.geoms[0]
    elif isinstance(geom, LineString) and list(geom.coords):
        return Point(geom.coords[0])
    elif geom.is_empty:
        raise IndexError(f"expected point at xs-river intersection got: {type(geom)} | {geom}")
    else:
        raise TypeError(f"expected point at xs-river intersection got: {type(geom)} | {geom}")


class RequireGeos:
    """Unsure."""

    def __init__(self, version):
        if version.count(".") != 2:
            raise ValueError("Version must be <major>.<minor>.<patch> format")
        self.version = tuple(int(x) for x in version.split("."))

    def __call__(self, func):
        """Call."""
        is_compatible = lib.geos_version >= self.version
        is_doc_build = os.environ.get("SPHINX_DOC_BUILD") == "1"  # set in docs/conf.py
        if is_compatible and not is_doc_build:
            return func  # return directly, do not change the docstring

        msg = "'{}' requires at least GEOS {}.{}.{}.".format(func.__name__, *self.version)
        if is_compatible:

            @wraps(func)
            def wrapped(*args, **kwargs):
                return func(*args, **kwargs)

        else:

            @wraps(func)
            def wrapped(*args, **kwargs):
                raise UnsupportedGEOSVersionError(msg)

        doc = wrapped.__doc__
        if doc:
            # Insert the message at the first double newline
            position = doc.find("\n\n") + 2
            # Figure out the indentation level
            indent = 0
            while True:
                if doc[position + indent] == " ":
                    indent += 1
                else:
                    break
            wrapped.__doc__ = doc.replace("\n\n", "\n\n{}.. note:: {}\n\n".format(" " * indent, msg), 1)

        return wrapped


def multithreading_enabled(func):
    """
    Prepare multithreading by setting the writable flags of object type ndarrays to False.

    NB: multithreading also requires the GIL to be released, which is done in
    the C extension (ufuncs.c).
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        array_args = [arg for arg in args if isinstance(arg, np.ndarray) and arg.dtype == object] + [
            arg
            for name, arg in kwargs.items()
            if name not in {"where", "out"} and isinstance(arg, np.ndarray) and arg.dtype == object
        ]
        old_flags = [arr.flags.writeable for arr in array_args]
        try:
            for arr in array_args:
                arr.flags.writeable = False
            return func(*args, **kwargs)
        finally:
            for arr, old_flag in zip(array_args, old_flags):
                arr.flags.writeable = old_flag

    return wrapped


@RequireGeos("3.7.0")
@multithreading_enabled
def reverse(geometry, **kwargs):
    """Return a copy of a Geometry with the order of coordinates reversed.

    If a Geometry is a polygon with interior rings, the interior rings are also
    reversed.

    Points are unchanged. None is returned where Geometry is None.

    Parameters
    ----------
    geometry : Geometry or array_like
    **kwargs
        See `NumPy ufunc docs <ufuncs.kwargs>` for other keyword arguments.

    See Also
    --------
    is_ccw : Checks if a Geometry is clockwise.

    Examples
    --------
    >>> from shapely import LineString, Polygon
    >>> reverse(LineString([(0, 0), (1, 2)]))
    <LINESTRING (1 2, 0 0)>
    >>> reverse(Polygon([(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)]))
    <POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))>
    >>> reverse(None) is None
    True
    """
    return lib.reverse(geometry, **kwargs)



"""
hecstac Package.

This package provides tools and utilities to create items, assets, and metadata from HEC models.
"""

from hecstac.hms.item import HMSModelItem
from hecstac.ras.item import RASModelItem
from hecstac.version import __version__



"""Features developed duirng R&D for consideration in the hecstac package."""

from pathlib import Path

from pystac import Asset, Item

from hecstac.common.logger import get_logger


def reorder_stac_assets(stac_item: Item) -> Item:
    """Sort assets alphabetically by key and move data assets to the top."""
    if not hasattr(stac_item, "assets") or not isinstance(stac_item.assets, dict):
        raise ValueError("STAC item must have an 'assets' attribute that is a dictionary.")

    data_assets = {k: v for k, v in sorted(stac_item.assets.items()) if hasattr(v, "roles") and ("data" in v.roles)}
    other_assets = {k: v for k, v in sorted(stac_item.assets.items()) if k not in data_assets}
    stac_item.assets = {**data_assets, **other_assets}

    return stac_item


def calibration_plots(stac_item: Item, plot_dir: str) -> Item:
    """Add existing calibration plots to a STAC item."""
    logger = get_logger(__name__)
    pngs = Path(plot_dir).rglob("*.png")
    for png in pngs:
        parent_dir = png.parent.parent.name
        asset_title = str(f"{parent_dir}/{png.stem}")
        logger.info(f"Adding asset: {asset_title}")
        new_asset = Asset(href=str(png), title=asset_title, media_type="image/png", roles=["thumbnail"])
        stac_item.add_asset(asset_title, new_asset)
    return stac_item



"""Geometry utils."""

from pyproj import CRS, Transformer
from shapely import Geometry
from shapely.ops import transform


def reproject_to_wgs84(geom: Geometry, crs: str) -> Geometry:
    """Convert geometry CRS to EPSG:4326 for stac item geometry."""
    pyproj_crs = CRS.from_user_input(crs)
    wgs_crs = CRS.from_authority("EPSG", "4326")
    if pyproj_crs != wgs_crs:
        transformer = Transformer.from_crs(pyproj_crs, wgs_crs, True)
        return transform(transformer.transform, geom)
    return geom


def read_crs_from_prj(prj_file: str) -> CRS:
    """Read CRS from a .prj file."""
    with open(prj_file, "r") as file:
        wkt = file.read()
    return CRS.from_wkt(wkt)
